/* nebula-client.c
 *
 * Copyright (C) 2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without
 * modifications, as long as this notice is preserved.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * Description:
 *   Simple command line tool which reads files and sends them to a Nebula server.
 */

#define _GNU_SOURCE 1

#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <zlib.h>

#include "md5.h"
#include "sha512.h"
#include "util.h"


// hmac stuff 
#define HMAC_HASH_SIZE	128	// for sha512
#define HMAC_BLOCK_SIZE	256	// for sha512
#define IPAD_VAL	0x36
#define OPAD_VAL	0x5C    


u_char	k[HMAC_BLOCK_SIZE];
u_char	k_ipad[HMAC_BLOCK_SIZE];
u_char	k_opad[HMAC_BLOCK_SIZE];


// read a line from a descriptor
int read_line(int socket, char *line, ssize_t len, int timeout) {
	int		read_chars = 0;
	int		rv = 0;

	memset(line, 0, len);

	/* read line without timeout */
	while (1) {
		if (read_chars >= len-1) {
			fprintf(stderr, "Error while reading from socket - Line exceeds buffer (%u bytes).\n", read_chars);
			exit(EXIT_FAILURE);
		}
		rv = recv(socket, &line[read_chars], 1, 0);
		if (rv == 0) return(read_chars);
		if (rv < 0) return(-1);
		if (line[read_chars] == '\n') {
			line[read_chars] = '\0';
			return(read_chars);
		}
		read_chars++;
	}
	return(0);
}


// calculate HMAC
char *hmac(u_char **msg, ssize_t len) {
	u_char	*inner, *outer;

	// append inner padding to message
	if ((*msg = realloc(*msg, len+HMAC_BLOCK_SIZE)) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %m.\n");
		return(NULL);
	}
	memcpy(*msg+len, k_ipad, HMAC_BLOCK_SIZE);

	// compute inner hash
	if ((inner = (u_char *) mem_sha512sum(*msg, len+HMAC_BLOCK_SIZE)) == NULL) {
		fprintf(stderr, "Error - Unable to compute inner HMAC SHA512 hash.\n");
		return(NULL);
	}

	// append outer padding to iner hash
	if ((inner = realloc(inner, HMAC_HASH_SIZE+HMAC_BLOCK_SIZE)) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %m.\n");
		free(inner);
		return(NULL);
	}
	memcpy(&inner[HMAC_HASH_SIZE], k_opad, HMAC_BLOCK_SIZE);

	// compute outer hash
	if ((outer = (u_char *) mem_sha512sum(inner, HMAC_HASH_SIZE+HMAC_BLOCK_SIZE)) == NULL)
		fprintf(stderr, "Error - Unable to compute outer HMAC SHA512 hash.\n");

	free(inner);
	return((char *) outer);
}


// print usage
void usage(const char* progname, const int exit_val) {
	printf("Usage: %s -c host [-p port] [-d directory] [-s secret] [-v] [file ...]\n", progname);
	exit(exit_val);
}


int main(int argc, char **argv) {
	u_char			verbose, time_sort, protocol, *cbuf, response[9];
	u_int16_t		hmac_port, port, send_len_short;
	u_int32_t		send_len_long, nonce, nebula_port;
	char			option, *dirname, *curfile, *md5sum, *sha512sum, *nebula_host, *secret;
	int			i, n, sock_fd, total_files, num_of_files;
	unsigned long		cbuf_len;
	ssize_t			bytes_read, total_bytes;
	struct hostent		*host;
	struct sockaddr_in	sock;
	struct dirent		*dir_entry, **namelist;
	DIR			*dirp;
	bstring			bstr;

	i			= 0;
	verbose			= 0;
	time_sort		= 0;
	cbuf_len		= 0;
	cbuf			= NULL;
	nebula_port		= 4712;
	nebula_host		= NULL;
	dirname			= NULL;
	dirp			= NULL;
	dir_entry		= NULL;
	secret			= NULL;

	protocol		= 6;		// always send tcp
	port			= 0;		// always send port 0


	// process args
	while((option = getopt(argc, argv, "c:d:hp:s:t?v")) > 0) {
		switch(option) {
			case 'c':
				nebula_host = optarg;
				break;
			case 'd':
				dirname = optarg;
				if ((dirp = opendir(optarg)) == NULL) {
					fprintf(stderr, "Error - Unable to open directory %s: %m.\n", optarg);
					exit(EXIT_FAILURE);
				}
				if (chdir(optarg) != 0) {
					fprintf(stderr, "Error - Unable to change into directory %s: %m.\n", optarg);
					exit(EXIT_FAILURE);
				}
				break;
			case 'p':
				nebula_port = atoi(optarg);
				if (!nebula_port || nebula_port > 65535) {
					fprintf(stderr, "Error - Invalid port.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 's':
				secret = optarg;
				break;
			case 't':
				time_sort = 1;
				break;
			case 'v':
				verbose++;
				break;
			case 'h':
			case '?':
				usage(argv[0], EXIT_SUCCESS);
			default:
				break;
		}
	}


	// initialize HMAC pads
	memset(k_ipad, IPAD_VAL, HMAC_BLOCK_SIZE);
	memset(k_opad, OPAD_VAL, HMAC_BLOCK_SIZE);
	if (secret) for (i=0; i<strlen(secret); i++) {
		k_ipad[i] ^= secret[i];
		k_opad[i] ^= secret[i];
	}


	if (dirname) {
		printf("Scanning directory... \n");
		if (time_sort) {
			if ((n = scandir(dirname, &namelist, regular_file, timesort)) == -1) {
				fprintf(stderr, "Error - Unable to read directory entries: %m.\n");
				exit(EXIT_FAILURE);
			}
		} else {
			if ((n = scandir(dirname, &namelist, regular_file, alphasort)) == -1) {
				fprintf(stderr, "Error - Unable to read directory entries: %m.\n");
				exit(EXIT_FAILURE);
			}
		}
		total_files = n;
	} else {
		if ((argc - optind) < 1) usage(argv[0], EXIT_FAILURE);
		i = optind;
		total_files = argc-i;
		if (verbose) printf("Processing files.\n");
	}


	// prepare socket
	if (!nebula_host && ((nebula_host = strdup("localhost")) == NULL)) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	if ((host = gethostbyname(nebula_host)) == NULL) {
		fprintf(stderr, "Error - Unable to resolve %s: %s.\n", nebula_host, strerror(errno));
		exit(EXIT_FAILURE);
	}


	// process files
	printf("Processing %u files.\n", total_files);
	for (n=0; n<total_files; n++) {
		curfile = (dirname ? namelist[n]->d_name : argv[n+i]);

		if (verbose == 1) printf("processing file %u\n", n);
		if (verbose > 1) printf("  processing file %s.\n", curfile);

		bstr = bstr_map(curfile);
		num_of_files++;

		// skip empty files
		if (!bstr.len) {
			bstr_unmap(bstr);
			continue;
		}

		/* create socket and connect to nebula server */
		if (!(sock_fd = socket(AF_INET, SOCK_STREAM, 0))) {
			fprintf(stderr, "Error - Unable to create socket: %m.\n");
			exit(EXIT_FAILURE);
		}

		memset(&sock, 0, sizeof(struct sockaddr_in));
		sock.sin_family	= AF_INET;
		sock.sin_addr	= *(struct in_addr*)host->h_addr;
		sock.sin_port	= htons(nebula_port);

		if (verbose > 1) printf("Connecting to %s:%u.\n", inet_ntoa(*(struct in_addr*)host->h_addr), nebula_port);
		if (connect(sock_fd, (struct sockaddr *) &sock, sizeof(struct sockaddr_in)) == -1) {
			fprintf(stderr, "Error - Unable to connect to %s:%u: %m\n",
				inet_ntoa(*(struct in_addr*)host->h_addr), nebula_port);
			close(sock_fd);
			exit(EXIT_FAILURE);
		}
		if (verbose > 1) printf("Connection to %s:%d established.\n",
			inet_ntoa(*(struct in_addr*)host->h_addr), nebula_port);



		// get nonce from server
		/* wait for incoming data, close connection on timeout */
		if (verbose > 1) printf("Reading nonce from server.\n");

		for (bytes_read = 1, total_bytes = 0; bytes_read && total_bytes < 4; total_bytes += bytes_read)
			bytes_read = read(sock_fd, &nonce+total_bytes, 4);

		if (bytes_read < 0) {
			fprintf(stderr, "Error - Unable to read from socket: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}
		nonce = ntohl(nonce);
		if (verbose > 1) printf("Nonce received.\n");


		// hash secret with nonce
		if (secret) {
				if ((cbuf = malloc(strlen(secret)+4)) == NULL) {
					fprintf(stderr, "Error - Unable to allocate memory: %m.\n");
					close(sock_fd);
					exit(EXIT_FAILURE);
				}
			memcpy(cbuf, secret, strlen(secret));
			memcpy(cbuf+strlen(secret), &nonce, 4);
			if ((sha512sum = mem_sha512sum(cbuf, strlen(secret)+4)) == NULL) {
				fprintf(stderr, "Error - Unable to hash secret.\n");
				close(sock_fd);
				exit(EXIT_FAILURE);
			}
			free(cbuf);
		} else {
			if ((sha512sum = mem_sha512sum((u_char *) &nonce, 4)) == NULL) {
				fprintf(stderr, "Error - Unable to hash nonce.\n");
				close(sock_fd);
				exit(EXIT_FAILURE);
			}
		}

		// send hashed secret or nonce
		if (write(sock_fd, sha512sum, 128) == -1) {
			fprintf(stderr, "Error - Writing to socket failed: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}
		free(sha512sum);

		// send md5 hash
		if ((md5sum = mem_md5sum(bstr.data, bstr.len)) == NULL) {
			fprintf(stderr, "Error - Calculating MD5 hash failed.\n");
			exit(EXIT_FAILURE);
		}
		if (write(sock_fd, md5sum, 32) == -1) {
			fprintf(stderr, "Error - Writing to socket failed: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}
		free(md5sum);


		if (!read_line(sock_fd, (char *) &response, 9, 10)) {
			printf("Warning - Nebula server did not respond within 10 seconds, skipping submission.\n");
			close(sock_fd);
			bstr_unmap(bstr);
			continue;
		}
		if (strncmp((char *) response, "KNOWN", 5) == 0) {
			if (verbose > 1) printf("Attack hash is already known to the server, skipping submission.\n");
			close(sock_fd);
			bstr_unmap(bstr);
			continue;
		}
		if (strncmp((char *) response, "UNKNOWN", 7)) {
			if (verbose) printf("Warning - Nebula server returned an invalid response, skipping submission.\n");
			close(sock_fd);
			bstr_unmap(bstr);
			continue;
		}
		if (verbose > 1) printf("Nebula server requested unknown attack, starting submission.\n");


		// send protocol 
		if (write(sock_fd, &protocol, 1) == -1) {
			fprintf(stderr, "Error - Writing to socket failed: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}

		// send port
		hmac_port = htons(port);
		if (write(sock_fd, &hmac_port, 2) == -1) {
			fprintf(stderr, "Error - Writing to socket failed: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}

		// send length of uncompressed attack
		send_len_long = htonl(bstr.len);
		if (write(sock_fd, &send_len_long, 4) == -1) {
			fprintf(stderr, "Error - Writing to socket failed: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}

		// compress attack
		if (verbose > 1) printf("Compressing attack data.\n");
		cbuf_len = (bstr.len * 1.1) + 12;
		if ((cbuf = calloc(1, cbuf_len)) == NULL) {
			fprintf(stderr, "Error - Unable to allocate memory: %m.\n");
			exit(EXIT_FAILURE);
		}
		if (verbose > 1) printf("Compressed data has %lu bytes.\n", cbuf_len);
		switch (compress(cbuf,
					&cbuf_len,
					bstr.data, bstr.len)) {
		case Z_OK:
			break;
		case Z_MEM_ERROR:
			fprintf(stderr, "Error - Cannot compress attack data: Out of memory.\n");
			exit(EXIT_FAILURE);
		case Z_BUF_ERROR:
			fprintf(stderr, "Error - Cannot compress attack data: Output buffer too small.\n");
			exit(EXIT_FAILURE);
		default:
			fprintf(stderr, "Error - Cannot compress attack data: Unknown error.\n");
			exit(EXIT_FAILURE);
		}
		if (verbose > 1) printf("Compressed data has %lu bytes.\n", (unsigned long) cbuf_len);

		// send length of compressed attack
		send_len_long = htonl(cbuf_len);
		if (write(sock_fd, &send_len_long, 4) == -1) {
			fprintf(stderr, "Error - Writing to socket failed: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}

		// send compressed attack
		if (write(sock_fd, cbuf, cbuf_len) == -1) {
			fprintf(stderr, "Error - Writing to socket failed: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}

		// append protocol and port to cattack for HMAC calculation
		if ((cbuf = realloc(cbuf, cbuf_len+3)) == NULL) {
			fprintf(stderr, "Error - Unable to allocate memory: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}
		memcpy(cbuf+cbuf_len, &protocol, 1);
		memcpy(cbuf+cbuf_len+1, &hmac_port, 2);

		if ((sha512sum = hmac(&cbuf, cbuf_len+3)) == NULL) {
			free(cbuf);
			close(sock_fd);
			exit(EXIT_FAILURE);
		}
		free(cbuf);

		// send length of HMAC
		send_len_short = htons(strlen(sha512sum));
		if (write(sock_fd, &send_len_short, 2) == -1) {
			fprintf(stderr, "Error - Writing to socket failed: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}

		// send HMAC
		if (write(sock_fd, sha512sum, strlen(sha512sum)) == -1) {
			fprintf(stderr, "Error - Writing to socket failed: %m.\n");
			close(sock_fd);
			exit(EXIT_FAILURE);
		}
		free(sha512sum);

		// wait for OK
		if (verbose > 1) printf("Attack sent, waiting for OK.\n");
		if (!read_line(sock_fd, (char *) response, 9, 10)) {
			printf("Warning - Nebula server did not respond within 10 seconds.\n");
		} else if (strlen((char *) response) != 2 || strncmp((char *) response, "OK", 2) != 0)
			printf("Warning - Invalid response from Nebula server: %s.\n", response);

		close(sock_fd);
		if (verbose > 1) printf("Submission complete.\n\n");

		bstr_unmap(bstr);
	}

	return(0);
}
