/* avl.c
 *
 * Copyright (C) 2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "avl.h"


void avl_free_subtree(avl_node *n) {
	if (n) {
		avl_free_subtree(n->left);
		avl_free_subtree(n->right);
		if (n->data) free(n->data);
		free(n);
	}
}

avl_tree *avl_tree_new(void) {
	avl_tree *t;

	if ((t = calloc(1, sizeof(avl_tree))) == NULL) {
		fprintf(stderr, "Could not allocate memory: %s.\n", strerror(errno));
		return(NULL);
	}
	return(t);
}


void avl_tree_free(avl_tree *t) {
	if (t == NULL) return;

	avl_free_subtree(t->root);
	free(t);
	return;
}


void avl_tree_reset(avl_tree *t) {
	avl_free_subtree(t->root);
	t->root		= NULL;
	t->size		= 0;
}


void avl_swl(avl_node** root) {
	avl_node*	a = *root;
	avl_node*	b = a->right;

	*root		= b;
	a->right	= b->left;
	b->left		= a;
}


void avl_swr(avl_node** root) {
	avl_node*	a = *root;
	avl_node*	b = a->left;

	*root		= b;
	a->left		= b->right;
	b->right	= a;
}


void avl_nasty(avl_node* root){
	switch (root->balance) {
	case -1:
		root->left->balance=0;
		root->right->balance=1;
		break;
	case 1:
		root->left->balance=-1;
		root->right->balance=0;
		break;
	case 0:
		root->left->balance=0;
		root->right->balance=0;
	}
	root->balance=0;
}


avl_node *avl_find(avl_tree *t, avl_node *n, const avl_node *a) {
	if (t == NULL || n == NULL || a == NULL) return(NULL);
	if (t->cmpfn(a->data, a->size, n->data, n->size) < 0) return(avl_find(t, n->left, a));
	if (t->cmpfn(a->data, a->size, n->data, n->size) > 0) return(avl_find(t, n->right, a));

	return(n);
}


int avl_insert(avl_tree *t, avl_node **n, avl_node *a) {
	if (!t) return(-1);

	if (!t->root) {
		t->root = a;
		return 1;
	}
   
	if (t->cmpfn((*n)->data, (*n)->size, a->data, a->size) > 0) {
	// insert into the left subtree
	if ((*n)->left) {
		if (avl_insert(t, &((*n)->left), a)) {
			switch ((*n)->balance--) {
			case 1:
				return 0;
			case 0:
				return 1;
			}
			if ((*n)->left->balance < 0) {
				avl_swr(n);
				(*n)->balance		= 0;
				(*n)->right->balance	= 0;
			} else {
				avl_swl(&((*n)->left));
				avl_swr(n);
				avl_nasty(*n);
			}
		}
		return 0;
	} else {
		(*n)->left = a;
		if ((*n)->balance--) return 0;
		return 1;
	}
   } else {
	// insert into the right subtree
	if ((*n)->right) {
		if (avl_insert(t, &((*n)->right), a)) {
			switch ((*n)->balance++) {
			case -1:
				return 0;
			case 0:
				return 1;
			}
			if ((*n)->right->balance > 0) {
				avl_swl(n);
				(*n)->balance		= 0;
				(*n)->left->balance	= 0;
			} else {
				avl_swr(&((*n)->right));
				avl_swl(n);
				avl_nasty(*n);
			}
		}
		return 0;
	} else {
		(*n)->right = a;
		if ((*n)->balance++) return 0;
		return 1;
	}
   }
}


void avl_subtree_print(avl_node *n) {
	if (n == NULL) return;
	avl_subtree_print(n->left);
	printf(" \033[1;33m%u\033[1;0m", *(u_int32_t*) n->data);
	avl_subtree_print(n->right);
	return;
}
