/* classify.c
 *
 * Copyright (C) 2007-2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <unistd.h>

#include "classify.h"
#include "cluster.h"
#include "nebula.h"
#include "queue.h"
#include "spamsum.h"
#include "session.h"
#include "sig.h"
#include "util.h"

#include <ctype.h>


// pthread wrapper function
void *pt_pollcq(void *param) {
	int			rv;
	qelem			*elem;

	for (;;) {
		if (repsem_wait(&sessions_semaphore) == -1) {
			fprintf(stderr, "Error - Unable to decrement semaphore: %s.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		pthread_mutex_lock(&classifyq_mutex);
		elem = queue_cuthead(classifyq);
		pthread_mutex_unlock(&classifyq_mutex);

		if (elem) {
			// call classification function
			if ((rv = classify((submission *) elem->data)) == -1) {
				fprintf(stderr, "Error - Unable to classify submission.\n");
				free(elem);
				return((void *) -1);
			} else if (rv == 0) {
				// known attack, do nothing
				free(elem);
				return((void *) 0);
			} else {
				// new attack
				free(elem->data);
				free(elem);
			}
		}
	}
	
	return((void *) 1);
}


// automated input classification (clustering)
int classify(submission *s) {
	trie_node	*t;
	hash		*tmp_hash;
	u_char		*tmpbuf;
	qelem		*cur_cqelem, *tmp_cqelem, *cur_hqelem, *tmp_hqelem, *qtail, *jqelem;
	double		score, max_score;

	qtail		= NULL;
	tmpbuf		= NULL;
	max_score	= 0.0;

	if (!s) {
		fprintf(stderr, "Error - Cannot classify empty session.\n");
		return(-1);
	}

	// create new element in md5sum trie
	pthread_rwlock_wrlock(&md5sum_trie_lock);
	t = trie_memins(&md5sum_trie, (u_char *) s->md5sum, strlen(s->md5sum), NULL);
	pthread_rwlock_unlock(&md5sum_trie_lock);

	if ((t->data = calloc(1, sizeof(hash))) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	((hash*)t->data)->hashlen	= 32;
	((hash*)t->data)->md5sum	= s->md5sum;

	if ((((hash*)t->data)->submission = calloc(1, sizeof(submission))) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: m.\n");
		exit(EXIT_FAILURE);
	}
	memcpy(((hash*)t->data)->submission, s, sizeof(submission));
	((hash*)t->data)->cnt++;

	// set spamsum hash
	if ((((hash*)t->data)->spamsum = spamsum(s->attack, s->attack_len, 0)) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (verbose > 1) printf("    MD5 hash is %s (%u instances)\n", ((hash*)t->data)->md5sum, ((hash*)t->data)->cnt);


	Q_RLOCK(&clusterq->lock);

	// connect all clusters within range
	for (cur_cqelem = clusterq->head; cur_cqelem; cur_cqelem = cur_cqelem->next) {
		for (cur_hqelem = ((cluster *)cur_cqelem->data)->hq->head; cur_hqelem; cur_hqelem = cur_hqelem->next) {
			if ((score = spamsum_match(((hash*)t->data)->spamsum, ((hash*)cur_hqelem->data)->spamsum)) >= sim_threshold) {
				if (!((hash*)t->data)->cl) {
					Q_ULOCK(&clusterq->lock);
					Q_WLOCK(&clusterq->lock);
					if ((qtail = add_entry_to_cluster((cluster *)cur_cqelem->data, (hash*)t->data, clusterhashq_max)) != NULL) {
						tmp_hash = qtail->data;

						pthread_rwlock_wrlock(&md5sum_trie_lock);
						trie_del_memstr(&md5sum_trie, (u_char *) tmp_hash->md5sum, strlen(tmp_hash->md5sum));
						pthread_rwlock_unlock(&md5sum_trie_lock);

						pthread_rwlock_wrlock(&spamsum_trie_lock);
						trie_del_memstr(&spamsum_trie, (u_char *) tmp_hash->spamsum, strlen(tmp_hash->spamsum));
						pthread_rwlock_unlock(&spamsum_trie_lock);

						hash_free(tmp_hash);
					}
					Q_ULOCK(&clusterq->lock);
					Q_RLOCK(&clusterq->lock);
					if (verbose) printf("    Cluster has now %lu elements (threshold is %lu).\n",
						(long unsigned) ((hash*)t->data)->cl->hq->size, ((hash*)t->data)->cl->threshold);
					break;
				} else {
					if ((cluster *) cur_cqelem->data != ((hash*)t->data)->cl)
						Q_ULOCK(&clusterq->lock);
						Q_WLOCK(&clusterq->lock);
						clusters_merge(clusterq, (cluster *) cur_cqelem->data, ((hash*)t->data)->cl);
						Q_ULOCK(&clusterq->lock);
						Q_RLOCK(&clusterq->lock);
					break;
				}
			} else if (score > max_score) max_score = score;
		}
	}
	Q_ULOCK(&clusterq->lock);

	// connect all outliers within range
	Q_RLOCK(&outlierq->lock);
	for (cur_hqelem = outlierq->head; cur_hqelem; cur_hqelem = cur_hqelem->next) {
		if ((score = spamsum_match(((hash*)t->data)->spamsum, ((hash*)cur_hqelem->data)->spamsum)) >= sim_threshold) {
			// unlink match from outlier list

			tmp_hqelem = cur_hqelem;
			cur_hqelem = cur_hqelem->next;
			if ((tmp_hash = queue_unlink(outlierq, tmp_hqelem)) == NULL) {
				fprintf(stderr, "Error - Unable to unlink outlier from queue.\n");
				exit(EXIT_FAILURE);
			}


			Q_WLOCK(&clusterq->lock);
			if (((hash*)t->data)->cl) {
				// add other outliers to cluster
				if ((qtail = add_entry_to_cluster(((hash*)t->data)->cl, tmp_hash, outlierq_max)) != NULL) {
					tmp_hash = qtail->data;

					pthread_rwlock_wrlock(&md5sum_trie_lock);
					trie_del_memstr(&md5sum_trie, (u_char *) tmp_hash->md5sum, strlen(tmp_hash->md5sum));
					pthread_rwlock_unlock(&md5sum_trie_lock);

					pthread_rwlock_wrlock(&spamsum_trie_lock);
					trie_del_memstr(&spamsum_trie, (u_char *) tmp_hash->spamsum, strlen(tmp_hash->spamsum));
					pthread_rwlock_unlock(&spamsum_trie_lock);

					hash_free(tmp_hash);
					free(qtail);
				} else if (verbose) printf("    Cluster has now %lu elements (threshold is %lu).\n",
					(long unsigned) ((hash*)t->data)->cl->hq->size, ((hash*)t->data)->cl->threshold);
			} else {
				// create new cluster of two outliers
				if ((tmp_cqelem = queue_ins(clusterq, create_cluster(((hash*)t->data), tmp_hash, clusterhashq_max),
					clusterq_max)) != NULL) {
					/* cluster queue is full, last element was dropped and must be free()d */
					
					/* first remove hashes from tries */
					for (tmp_hqelem = ((cluster *)tmp_cqelem->data)->hq->head; tmp_hqelem; tmp_hqelem=tmp_hqelem->next) {

						pthread_rwlock_wrlock(&md5sum_trie_lock);
						trie_del_memstr(&md5sum_trie, (u_char *) ((hash *)tmp_hqelem->data)->md5sum,
							strlen(((hash *)tmp_hqelem->data)->md5sum));
						pthread_rwlock_unlock(&md5sum_trie_lock);

						pthread_rwlock_wrlock(&spamsum_trie_lock);
						trie_del_memstr(&spamsum_trie, (u_char *) ((hash *)tmp_hqelem->data)->spamsum,
							strlen(((hash *)tmp_hqelem->data)->spamsum));
						pthread_rwlock_unlock(&spamsum_trie_lock);
					}
					/* now free cluster */
					cluster_free((cluster *)tmp_cqelem->data);
					free(tmp_cqelem);
				}
				/* set cluster element's parent pointer to cluster queue head
				 * we need this to be able to unlink a cluster from the queue */
				((cluster *) clusterq->head->data)->parent = clusterq->head;

				if (verbose) printf("    New Cluster created.\n");
			}
			Q_ULOCK(&clusterq->lock);
		} else if (score > max_score) max_score = score;
		if (!cur_hqelem) break;
	}
	Q_ULOCK(&outlierq->lock);

		
	if (verbose > 1) printf("    Spamsum is %s (%u instances)\n", ((hash*)t->data)->spamsum, ((hash*)t->data)->cnt);

	if (!((hash*)t->data)->cl) {
		// insert outlier into queue
		if (outlierq->size >= outlierq_max) {
			tmp_hash = queue_unlink(outlierq, outlierq->tail);

			pthread_rwlock_wrlock(&md5sum_trie_lock);
			trie_del_memstr(&md5sum_trie, (u_char *) tmp_hash->md5sum, strlen(tmp_hash->md5sum));
			pthread_rwlock_unlock(&md5sum_trie_lock);

			pthread_rwlock_wrlock(&spamsum_trie_lock);
			trie_del_memstr(&spamsum_trie, (u_char *) tmp_hash->spamsum, strlen(tmp_hash->spamsum));
			pthread_rwlock_unlock(&spamsum_trie_lock);

			hash_free(tmp_hash);
			free(qtail);
		}
		queue_ins(outlierq, t->data, outlierq_max);

		if (verbose) printf("    Input added to outlier queue (maximum score: %.1f, outlier queue size: %lu).\n",
			max_score, (long unsigned) outlierq->size);
	}

	pthread_mutex_lock(&siggenq_mutex);
	// check for signature generation criterion
	if ((((hash*)t->data)->cl) && ((hash*)t->data)->cl->hq->size >= ((hash*)t->data)->cl->threshold) {
		printf("[=] Cluster size (%lu) hit threshold (%lu), generating signature.\n",
			(long unsigned) ((hash*)t->data)->cl->hq->size, ((hash*)t->data)->cl->threshold);

		// increment size threshold
		if  (((long unsigned) ((hash*)t->data)->cl->hq->size) > ((hash*)t->data)->cl->threshold)
			((hash*)t->data)->cl->threshold = ((long unsigned) ((hash*)t->data)->cl->hq->size);
		((hash*)t->data)->cl->threshold *= 1.2;

		// make sure every cluster gets queued only once
		for (jqelem = siggenq->head; jqelem && jqelem->data != ((hash*)t->data)->cl; jqelem = jqelem->next);

		// generate signature for this cluster
		if (!jqelem || jqelem->data != ((hash*)t->data)->cl) {
			// append cluster to signature generation job queue
			if (queue_append(siggenq, (void *) (void *) ((hash*)t->data)->cl) == NULL) {
				fprintf(stderr, "Error - Unable to queue cluster for signature generation.\n");
				exit(EXIT_FAILURE);
			} 

			// wake up a signature generation thread
			if (repsem_post(&siggen_semaphore) == -1) {
				fprintf(stderr, "Error - Unable to increment semaphore: %s.\n", strerror(errno));
				exit(EXIT_FAILURE);
			}
		} else {
			printf("[=] Cluster is already queued for signature generation.\n");
		}
	}
	pthread_mutex_unlock(&siggenq_mutex);

	return(1);
}
