/* cluster.c
 *
 * Copyright (C) 2007 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <stdio.h>
#include <string.h>

#include "cluster.h"
#include "nebula.h"
#include "queue.h"


/* create new cluster */
inline cluster *cluster_new(void) {
	cluster *new;

	/* create new cluster */
	if ((new = calloc(1, sizeof(cluster))) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	if (pthread_rwlock_init(&new->lock, NULL) != 0) {
		fprintf(stderr, "Error - Unable to initialize thread lock: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	/* create element queue */
	if ((new->hq = calloc(1, sizeof(queue))) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %m.\n");
		exit(EXIT_FAILURE);
	}

	// initialize size threshold
	new->threshold = initial_threshold;

	return(new);
}


/* delete cluster */
void cluster_free(void *cl) {
	if (!cl) return;

	if (lock_mutex) CL_WLOCK(cl);
	queue_free(((cluster *)cl)->hq, hash_free);
	if (lock_mutex) CL_ULOCK(cl);
	free(cl);

	return;
}


qelem *add_entry_to_cluster(cluster *cl, hash *entry, const ssize_t max_size) {
	qelem	*new;

	CL_WLOCK(cl);

	// set cluster pointer in hash struct
	entry->cl = cl;

	// add hash as new head to queue
	if ((new = queue_prepend(cl->hq, entry)) == NULL) {
		fprintf(stderr, "Error - Could not add entry to cluster.\n");
		exit(EXIT_FAILURE);
	}

	if (max_size && cl->hq->size > max_size) {
		// max queue size reached, drop and return tail element
		if (verbose > 1) fprintf(stderr, "    Maximum queue size reached, dropping tail.\n");
		CL_ULOCK(cl);
		return(queue_cuttail(cl->hq));
	}

	CL_ULOCK(cl);
	return(NULL);
}


cluster *create_cluster(hash *l1, hash *l2, const ssize_t max_size) {
	cluster *cl = cluster_new();

	add_entry_to_cluster(cl, l1, max_size);
	add_entry_to_cluster(cl, l2, max_size);

	return(cl);
}


cluster *clusters_merge(queue *q, cluster *dst, cluster *src) {
		qelem	*entry;

		CL_WLOCK(dst);
		CL_WLOCK(src);

		// concatenate hash queues
		dst->hq->tail->next = src->hq->head;
		dst->hq->tail = src->hq->tail;
		dst->hq->size += src->hq->size;

		// calculate new thresholds
		if (src->hq->size + dst->hq->size >= max(src->threshold, dst->threshold))
			dst->threshold += src->threshold;
		else
			dst->threshold = max(src->threshold, dst->threshold);

		// let elements of src point to dst
		for (entry = src->hq->head; entry; entry = entry->next) ((hash*)entry->data)->cl = dst;
		src->hq->head = src->hq->tail = NULL;

		CL_ULOCK(dst);
		CL_ULOCK(src);

		// unlink and free old cluster
		queue_free(src->hq, NULL);
		entry = queue_unlink(q, src->parent);
		cluster_free(entry->data);
		free(entry);

		return(dst);
}
