/* lca.c
 *
 * Copyright (C) 2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * This code is based on Gusfield's strmat package.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lca.h"
#include "stree.h"


static int compute_I_and_L(lcatbl *lca, stree *tree, stnode *node);
static void compute_A(lcatbl *lca, stree *tree, stnode *node, u_int32_t Amask);
static u_int32_t h(u_int32_t number);
static u_int32_t MSB(u_int32_t number);


static u_int32_t msb_table[256], lsb_table[256];
static int init_flag = 0;


static void init_tables(void) {
	int i, j, lsb, msb, mask;

	j	= 1;
	msb	= -1;

	for (i=1; i < 256; i++) {
		if (i == j) {
			j <<= 1;
			msb++;
		}
		msb_table[i] = msb;

		for (lsb=0, mask=1; !(mask & i); lsb++, mask<<=1);

		lsb_table[i] = lsb;
	}
	init_flag	= 1;
}


static u_int32_t h(u_int32_t number) {
	if (!init_flag) init_tables();

	if (number & 0xff) return lsb_table[number & 0xff];
	number >>= 8;
	if (number & 0xff) return 8 + lsb_table[number & 0xff];
	number >>= 8;
	if (number & 0xff) return 16 + lsb_table[number & 0xff];
	number >>= 8;
	if (number & 0xff) return 24 + lsb_table[number & 0xff];

	return 32;
}


static u_int32_t MSB(u_int32_t number) {
	if (!init_flag) init_tables();

	if (number & (0xff << 24)) return 24 + msb_table[number >> 24];
	if (number & (0xff << 16)) return 16 + msb_table[number >> 16];
	if (number & (0xff << 8)) return 8 + msb_table[number >> 8];
	if (number & 0xff) return msb_table[number];

	return 32;
}
  

lcatbl *lca_prep(stree *tree) {
	u_int32_t	num_nodes;
	lcatbl	*lca;

	if (tree == NULL) return NULL;

	if ((lca = calloc(1, sizeof(lcatbl))) == NULL) return NULL;

	lca->tree	= tree;
	num_nodes	= tree->size + 1;

	if (((lca->I = calloc(num_nodes, sizeof(u_int32_t))) == NULL) ||
	    ((lca->A = calloc(num_nodes, sizeof(u_int32_t))) == NULL) ||
	    ((lca->L = calloc(num_nodes, sizeof(stnode))) == NULL)) {
		lca_free(lca);
		return NULL;
	}

	/* Compute the I and L values, then compute the A values. */
	compute_I_and_L(lca, tree, tree->root);
	compute_A(lca, tree, tree->root, 0);

	return lca;
}


static int compute_I_and_L(lcatbl *lca, stree *tree, stnode *node) {
	u_int32_t	Ival, Imax;
	stnode		*child;

	/* Find the node with the maximum I value in the subtree.  */
	Imax = node->id;
	for (child = node->sons; child; child = child->right_sibling) {
		Ival = compute_I_and_L(lca, tree, child);
		if(h(Ival) > h(Imax)) Imax = Ival;
	}

	lca->I[node->id]	= Imax;
	lca->L[Imax]		= *node;    /* will be overwritten by the highest node in run */

	return Imax;
}


static void compute_A(lcatbl *lca, stree *tree, stnode *node, u_int32_t Amask) {
	stnode		*child;

	Amask |= 1 << h(lca->I[node->id]);
	lca->A[node->id] = Amask;

	for (child = node->sons; child; child = child->right_sibling) compute_A(lca, tree, child, Amask);
}


stnode *lca_lookup(lcatbl *lca, stnode *x, stnode *y) {
	u_int32_t	k, b, j, l, Iw, mask, *I, *A;
	stnode		*L, *w, *xbar, *ybar;
	stree		*tree;

	if (lca == NULL || x == NULL || y == NULL) return NULL;

	tree	= lca->tree;
	I	= lca->I;
	A	= lca->A;
	L	= lca->L;

	// steps 1 and 2
	k	= MSB(I[x->id] ^ I[y->id]);
	mask	= ~0 << (k + 1);
	b	= (I[x->id] & mask) | (1 << k);

	mask	= ~0 << h(b);
	j	= h( (A[x->id] & A[y->id]) & mask );

	// step 3
	l = h(A[x->id]);
	if (l == j) xbar = x;
	else {
		mask	= ~(~0 << j);	// the bit-complement of setting bits j..32
		k	= MSB(A[x->id] & mask);

		mask	= ~0 << (k + 1);
		Iw	= (I[x->id] & mask) | (1 << k);
		w	= &L[Iw];
		xbar	= w->father;
	}

	// step 4
	l = h(A[y->id]);
	if (l == j) ybar = y;
	else {
		mask	= ~(~0 << j);
		k	= MSB(A[y->id] & mask);

		mask	= ~0 << (k + 1);
		Iw	= (I[y->id] & mask) | (1 << k);
		w	= &L[Iw];
		ybar	= w->father;
	}

	return(xbar->id < ybar->id ? xbar : ybar);
}


void lca_free(lcatbl *lca) {
	if (lca == NULL) return;

	if (lca->I != NULL) free(lca->I);
	if (lca->A != NULL) free(lca->A);
	if (lca->L != NULL) free(lca->L);

	free(lca);
}
