/* nebula.c
 *
 * Copyright (C) 2007-2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <poll.h>
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <string.h>
#include <unistd.h>

#include "classify.h"
#include "cluster.h"
#include "md5.h"
#include "nebula.h"
#include "net.h"
#include "session.h"
#include "sig.h"
#include "signals.h"
#include "spamsum.h"
#include "util.h"

#ifndef VERSION
#define VERSION "(unknown version)"
#endif


void usage(const char* progname, const int exit_val) {
	printf("Usage: %s "
		" -a <filename>\t append new snort signatures to this file\n"
		"\t\t -C <size>\t cluster queue size\n"
		"\t\t -c <similarity> cluster criterion (a similarity measure in percent)\n"
		"\t\t -d\t\t daemonize\n"
		"\t\t -e\t\t minimum substring entropy\n"
		"\t\t -g\t\t number of signature generation threads\n"
		"\t\t -h\t\t this help\n"
		"\t\t -E <size>\t cluster element queue size\n"
		"\t\t -i <sid>\t initial snort signature ID\n"
		"\t\t -l\t\t minimum substring length\n"
		"\t\t -O <size>\t outlier queue size\n"
		"\t\t -p <port>\t listen on this port\n"
		"\t\t -r <snort pid>\t send a SIGHUP to this process ID after a new rule was generated\n"
		"\t\t -s <secret>\t secret string for use in submissions\n"
		"\t\t -t <threshold>\t initial cluster element threshold\n"
		"\t\t -v\t\t be verbose\n"
		"\t\t\t\t [files ...]\n",
		progname);

	exit(exit_val);
}


int main(int argc, char *argv[]) {
	int		i, qsize, port, listen_fd, rv, gthread_no; 
	u_char		daemonize;
	char		option;
	pthread_t	ntid;
	submission	*tmp_submission;
	pthread_attr_t	ptattr;
	struct rlimit	rlim;


	memset(pfdset, 0, sizeof(struct pollfd) * (POLLFD_MAX_SET_SIZE+1));
	memset(s, 0, sizeof(submission) * (POLLFD_MAX_SET_SIZE+1));
	pollfd_set_size		= 0;

	outlierq		= NULL;		// outlier queue
	clusterq		= NULL;		// cluster queue

	classifyq		= NULL;		// classification job queue
	siggenq			= NULL;		// signature generation job queue

	i			= 0;
	qsize			= 0;

	lock_mutex		= 1;		// this flag is for avoiding deadlocks during shutdown
						// if set threads will lock mutexes
						// a shutdown routine should clean it before further operations

	// default values for command line parameters
	rules_file		= NULL;		// a: don't append snort rules to a file
	sim_threshold		= 70.0;		// c: 70% similarity as cluster criterion
	clusterq_max		= 5000;		// C: cluster queue size
	daemonize		= 0;		// d: don't daemonize
	min_sstr_ent		= 0;		// e: minimum substring entropy
	clusterhashq_max	= 500000;	// E: cluster element queue size
	gthread_no		= 1;		// g: spawn 1 signature generation thread
	global_sid		= 2000000;	// i: initial snort ID (fist signature gets global_sid+1)
	min_sstr_len		= 0;		// l: minimum substring length (0 == infinite) 
	outlierq_max		= 500000;	// O: outlier element queue size
	port			= 4712;		// p: listen on port 4712/tcp
	snort_pid		= 0;		// r: PID of snort process for ruleset reloads (0 == off)
	secret			= NULL;		// s: don't use a secret
	initial_threshold	= 30;		// t: initial cluster element threshold for signature generation
	verbose			= 0;		// v: don't be verbose

	// prepare tries for hash lookups
	memset(&md5sum_trie, 0, sizeof(trie_node));
	memset(&spamsum_trie, 0, sizeof(trie_node));
	memset(&sighash_trie, 0, sizeof(trie_node));

	// initialize thread locks
	if (pthread_rwlock_init(&md5sum_trie_lock, NULL) ||
	    pthread_rwlock_init(&spamsum_trie_lock, NULL) ||
	    pthread_rwlock_init(&sighash_trie_lock, NULL) ||
	    pthread_rwlock_init(&sidlock, NULL) ||
	    pthread_mutex_init(&classifyq_mutex, NULL) ||
	    pthread_mutex_init(&siggenq_mutex, NULL) ||
	    pthread_mutex_init(&siggen_mutex, NULL) ||
	    pthread_mutex_init(&sigwrite_mutex, NULL)) {
		fprintf(stderr, "Error - Unable to initialize thread lock: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}


	// initialize semaphore
	if ((repsem_init(&sessions_semaphore, 0, 0) == -1) ||
	    (repsem_init(&sessions_semaphore, 0, 0) == -1)) {
		fprintf(stderr, "Error - Unable to initialize semaphore: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	printf("\n  Nebula %s Copyright (C) 2007-2008 Tillmann Werner <tillmann.werner@gmx.de>\n\n", VERSION);

	// process args
	while((option = getopt(argc, argv, "a:c:C:de:E:g:hi:l:O:p:r:s:t:v?")) > 0) {
		switch(option) {
			case 'a':
				rules_file = optarg;
				break;
			case 'c':
				sim_threshold = atof(optarg);
				if ((sim_threshold < 0) || (sim_threshold > 100)) {
					fprintf(stderr, "Error - Cluster radius must be a value between 0 and 100 (in percent).\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 'C':
				clusterq_max = atoi(optarg);
				if (clusterq_max < 1) {
					fprintf(stderr, "Error - Cluster queue size must be a non-negative value.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 'd':
				daemonize = 1;
				break;
			case 'e':
				min_sstr_ent = atoi(optarg);
				if (min_sstr_ent < 0) {
					fprintf(stderr, "Error - Minimum substring entropy cannot be negative.\n");
					exit(EXIT_FAILURE);
				} else if (min_sstr_ent > 8) {
					fprintf(stderr, "Error - Minimum substring entropy cannot be bigger than 8.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 'E':
				clusterhashq_max = atoi(optarg);
				if (clusterhashq_max < 2) {
					fprintf(stderr, "Error - Cluster element queue size must be at least 2.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 'g':
				gthread_no = atoi(optarg);
				if (gthread_no < 1) {
					fprintf(stderr, "Error - Need at least one signature generation thread.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 'i':
				global_sid = strtoul(optarg, NULL, 10);
				if (!global_sid) {
					fprintf(stderr, "Error - Invalid initial snort signature ID.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 'l':
				min_sstr_len = atoi(optarg);
				if (min_sstr_len < 0) {
					fprintf(stderr, "Error - Minimum substring length cannot be negative.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 'O':
				outlierq_max = atoi(optarg);
				if (outlierq_max < 1) {
					fprintf(stderr, "Error - Outlier queue size must be a non-negative value.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 'p':
				port = atoi(optarg);
				if (!port || port > 65535) {
					fprintf(stderr, "Error - Invalid port.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 'r':
				snort_pid = atoi(optarg);
				if (!snort_pid) {
					fprintf(stderr, "Error - Invalid process ID.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 's':
				secret = optarg;
				break;
			case 't':
				initial_threshold = strtoul(optarg, NULL, 10);
				if (initial_threshold < 2) {
					fprintf(stderr, "Error - Invalid initial threshold.\n");
					exit(EXIT_FAILURE);
				}
				break;
			case 'v':
				verbose++;
				break;
			case 'h':
			case '?':
				usage(argv[0], EXIT_SUCCESS);
			default:
				break;
		}
	}

	// set up signal handling
	set_signal_handlers();


	// spawn classification thread
	if (pthread_attr_init(&ptattr) || pthread_attr_setdetachstate(&ptattr, PTHREAD_CREATE_DETACHED)) {
		fprintf(stderr, "Error - Unable to initialize thread attributes: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (pthread_create(&ntid, &ptattr, pt_pollcq, NULL)) {
		fprintf(stderr, "Error - Cannot create clustering thread: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	pthread_attr_destroy(&ptattr);


	// spawn signature generation thread(s)
	for (; gthread_no; gthread_no--) {
		if (verbose > 1) printf("  Signature generation thread spawned.\n");
		if (pthread_attr_init(&ptattr) || pthread_attr_setdetachstate(&ptattr, PTHREAD_CREATE_DETACHED)) {
			fprintf(stderr, "Error - Unable to initialize thread attributes: %s.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		if (pthread_create(&ntid, &ptattr, pt_pollgq, NULL)) {
			fprintf(stderr, "Error - Cannot create signature generation thread: %s.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		pthread_attr_destroy(&ptattr);
	}
	if (verbose > 1) putchar('\n');


	// increase maximum number of open files if necessary (only needed for submission backlog queue)
	memset(&rlim, 0, sizeof(struct rlimit));
	if (getrlimit(RLIMIT_NOFILE, &rlim) == -1) {
		fprintf(stderr, "Error - Unable to get maximum number of open files: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (rlim.rlim_max < POLLFD_MAX_SET_SIZE+1) {
		rlim.rlim_max = rlim.rlim_cur = POLLFD_MAX_SET_SIZE+1;
		if (setrlimit(RLIMIT_NOFILE, &rlim) == -1) {
			fprintf(stderr, "Error - Unable to increase maximum number of open files: %s.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		if (verbose) printf("  Maximum number of open files increased to %u.\n", (unsigned int) rlim.rlim_cur);
	}


	// submissions are secured with HMAC, initialize HMAC pads
	memset(k_ipad, IPAD_VAL, HMAC_BLOCK_SIZE);
	memset(k_opad, OPAD_VAL, HMAC_BLOCK_SIZE);
	if (secret) for (i=0; i<strlen(secret); i++) {
		k_ipad[i] ^= secret[i];
		k_opad[i] ^= secret[i];
	}


	// initialize queues
	outlierq	= queue_new();
	clusterq	= queue_new();
	classifyq	= queue_new();
	siggenq		= queue_new();


	// bind to port
	listen_fd = net_listen(port);


	// inform about configuration
	if (!secret) {
		printf("  Warning - No submission secret given.\n");
	} else if (verbose) printf("  Submission secret: %s\n", secret);

	if (rules_file && verbose)
		printf("  Appending new rules to %s.\n", rules_file);
	if (snort_pid && verbose)
		printf("  Forcing snort (process ID %u) reload on new rules.\n", snort_pid);

	if (verbose) {
		printf("  Initial snort signature ID: %u\n", global_sid);
		printf("  Initial cluster size threshold: %lu\n", initial_threshold);
		printf("  Cluster criterion (minimal similarity): %.1f percent\n", sim_threshold);
		printf("  Accepting submissions on port %u/tcp.\n", port);
	}
	printf("\n[*] Ready.\n");
	fflush(stdout);


	// process incoming connections
	LISTEN_SOCK.fd		= listen_fd;
	LISTEN_SOCK.events	= POLLIN;
	for(;;) {
		switch (rv = poll(pfdset, pollfd_set_size+1, -1)) {
		case -1:
			if (errno == EINTR) break;
			fprintf(stderr, "Error - Unable to poll sockets: %s.\n", strerror(errno));
			exit(1);
		case  0:
			break;
		default:
			// check listen_fd

			if (LISTEN_SOCK.revents) {
				if (LISTEN_SOCK.revents & POLLERR) {
					fprintf(stderr, "Error - Unable to poll listening socket: %s.\n", strerror(errno));
					exit(EXIT_FAILURE);
				} else if (LISTEN_SOCK.revents & POLLHUP) {
					fprintf(stderr, "Error - Listening socket hangup.\n");
					exit(EXIT_FAILURE);
				} else if (LISTEN_SOCK.revents & POLLNVAL) {
					fprintf(stderr, "Error - Listening socket descriptor is invalid.\n");
					exit(EXIT_FAILURE);
				}

				if ((LISTEN_SOCK.revents & POLLIN) && pollfd_set_size < POLLFD_MAX_SET_SIZE) {
					// incoming connection, accept it and place fd in pollfd set
					pollfd_set_size++;

					pfdset[pollfd_set_size].fd	= net_accept(LISTEN_SOCK.fd);
					pfdset[pollfd_set_size].events	= POLLOUT;

					memset(&s[pollfd_set_size], 0, sizeof(submission));

					if (verbose > 1) printf("[>] Connection accepted: %d.\n", pfdset[pollfd_set_size].fd);
				}
			}
			for (i=1; i<=pollfd_set_size; i++) {
				if (!pfdset[i].revents) continue;

				if (pfdset[i].revents & POLLOUT) {
					switch (session_handle_data(&pfdset[i], &s[i])) {
					case -1:
						fprintf(stderr, "Error - Invalid submission state. Terminating session.\n");
						session_reset(&s[i], i);
						break;
					case 0:
						session_reset(&s[i], i);
						break;
					default:
						pfdset[i].events = POLLIN;
						break;
					}
				} else if (pfdset[i].revents & POLLIN) {
					switch (session_handle_data(&pfdset[i], &s[i])) {
					case 2:
						// need more data to complete session
						break;
					case 1:
						// submission complete, create clustering thread
						if ((tmp_submission = calloc(1, sizeof(submission))) == NULL) {
							fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
							exit(EXIT_FAILURE);
						}
						memcpy(tmp_submission, &s[i], sizeof(submission));
						memset(&s[i], 0, sizeof(submission));
						
						// add session to classify queue
						pthread_mutex_lock(&classifyq_mutex);
						if (queue_append(classifyq, (void *) tmp_submission) == NULL) {
							fprintf(stderr, "Error - Unable to queue session for classification.\n");
							exit(EXIT_FAILURE);
						} 
						pthread_mutex_unlock(&classifyq_mutex);

						// wake up a classification thread
						if (repsem_post(&sessions_semaphore) == -1) {
							fprintf(stderr, "Error - Unable to increment semaphore: %s.\n", strerror(errno));
							exit(EXIT_FAILURE);
						}

						session_reset(&s[i], i);
						break;
					case 0:
						// connection closed
						session_reset(&s[i], i);
						break;
					case -1:
						// program error
						fprintf(stderr, "Error - Terminating session.\n");
						session_reset(&s[i], i);
						break;
					case -2:
						// connection state error
						fprintf(stderr, "Error - Invalid connection state.\n");
						session_reset(&s[i], i);
						break;
					default:
						break;
					}
				} else if (pfdset[i].revents & POLLERR && errno != EINTR) {
					fprintf(stderr, "Error - Unable to poll socket: %s.\n", strerror(errno));
					session_reset(&s[i], i);
				} else if (pfdset[i].revents & POLLHUP) {
					fprintf(stderr, "Error - Socket hangup.\n");
					session_reset(&s[i], i);
				} else if (pfdset[i].revents & POLLNVAL) {
					fprintf(stderr, "Error - Socket descriptor is invalid.\n");
					session_reset(&s[i], i);
				}
			}
		}
	}

	// never reached
	cleanup();
	return(EXIT_SUCCESS);
}
