/* sem.c
 *
 * Copyright (C) 2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <errno.h>

#include "sem.h"


#define REPSEM_VALUE_MAX 32767


int repsem_init(repsem_t *sem, int pshared, int value) {
	int ret;

	if (value > REPSEM_VALUE_MAX) {
		errno = EINVAL;
		return -1;
	}

	if (pshared) {
		errno = ENOSYS;
		return -1;
	}

	sem->value = value;

	if ((ret = pthread_mutex_init(&sem->mutex, NULL)) < 0) return -1;
	if ((ret = pthread_cond_init(&sem->cond, NULL)) < 0) return -1;

	return 0;
}


int repsem_post(repsem_t *sem) {
	if (pthread_mutex_lock(&sem->mutex) < 0) return -1;

	if (sem->value == REPSEM_VALUE_MAX) {
		errno = ERANGE;
		return -1;
	}
	sem->value++;

	if (pthread_cond_signal(&sem->cond) < 0) {
		pthread_mutex_unlock(&sem->mutex);
		return -1;
	}

	if (pthread_mutex_unlock(&sem->mutex) < 0) return -1;

	return 0;
}


int repsem_wait(repsem_t *sem) {
	if (pthread_mutex_lock(&sem->mutex) < 0) return -1;

	while(sem->value == 0) pthread_cond_wait(&sem->cond,&sem->mutex);
	sem->value--;

	if (pthread_mutex_unlock(&sem->mutex) < 0) return -1;

	return 0;
}


int repsem_trywait(repsem_t *sem) {
	if (pthread_mutex_lock(&sem->mutex) < 0) return -1;

	if (sem->value == 0) {
		if (pthread_mutex_unlock(&sem->mutex) < 0) return -1;
		errno = EAGAIN;
		return -1;
	}

	sem->value--;

	if (pthread_mutex_unlock(&sem->mutex) < 0) return -1;

	return 0;
}


int repsem_getvalue(repsem_t *sem, int *sval) {
	*sval = sem->value;
	return 0;
}


int repsem_destroy(repsem_t *sem) {
	int ret;

	ret = pthread_cond_destroy(&sem->cond);
	if (ret < 0) return -1;

	ret = pthread_mutex_destroy(&sem->mutex);
	if (ret < 0) return -1;

	return 0;
}
