/* session.c
 *
 * Copyright (C) 2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <arpa/inet.h>
#include <errno.h>
#include <poll.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <zlib.h>

#include "classify.h"
#include "md5.h"
#include "nebula.h"
#include "session.h"
#include "sha512.h"
#include "util.h"


void session_reset(submission *subm, int session_number) {
	if (!subm) return;

	free(subm->cattack);
	free(subm->attack);
	free(subm->md5sum);
	free(subm->hmac);
	memset(subm, 0, sizeof(submission));

	if (session_number > 0) {
		if (pfdset[session_number].fd >= 0) close(pfdset[session_number].fd);
		if (verbose > 1) printf("[<] Connection terminated: %d.\n", pfdset[session_number].fd);

		// shrink pollfd set and adjust session array
		memmove(&pfdset[session_number], &pfdset[session_number+1], sizeof(struct pollfd) * (pollfd_set_size-session_number));
		memmove(&s[session_number], &s[session_number+1], sizeof(submission) * (pollfd_set_size-session_number));

		memset(&pfdset[pollfd_set_size], 0, sizeof(struct pollfd));
		memset(&s[pollfd_set_size], 0, sizeof(submission));

		pollfd_set_size--;
	}

	fflush(stdout);
	return;
}


char *session_hmac(u_char **msg, ssize_t len) {
	u_char	*inner, *outer;

	// append inner padding to message
	if ((*msg = realloc(*msg, len+HMAC_BLOCK_SIZE)) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		return(NULL);
	}
	memcpy(*msg+len, k_ipad, HMAC_BLOCK_SIZE);

	// compute inner hash
	if ((inner = (u_char *) mem_sha512sum(*msg, len+HMAC_BLOCK_SIZE)) == NULL) {
		fprintf(stderr, "Error - Unable to compute inner HMAC SHA512 hash.\n");
		return(NULL);
	}

	// append outer padding to iner hash
	if ((inner = realloc(inner, HMAC_HASH_SIZE+HMAC_BLOCK_SIZE)) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		free(inner);
		return(NULL);
	}
	memcpy(&inner[HMAC_HASH_SIZE], k_opad, HMAC_BLOCK_SIZE);

	// compute outer hash
	if ((outer = (u_char *) mem_sha512sum(inner, HMAC_HASH_SIZE+HMAC_BLOCK_SIZE)) == NULL)
		fprintf(stderr, "Error - Unable to compute outer HMAC SHA512 hash.\n");

	free(inner);
	return((char *) outer);
}


int session_read_data(int fd, submission *s, u_char *buf, ssize_t len) {
	int	bytes_read;

	if ((bytes_read = read(fd, buf + s->bytes_read, len - s->bytes_read)) == -1) {
		fprintf(stderr, "Error - Unable to read from socket: %s.\n", strerror(errno));
		return(-1);
	}

	s->bytes_read += bytes_read;

	return(s->bytes_read);
}


int session_handle_data(struct pollfd *pfd, submission *s) {
	int		rv;
	u_int16_t	hmac_port;
	u_int32_t	nonce;
	char		*md5sum;
	char		*sha512sum;
	char		*secret_nonce;
	trie_node	*t;


	if (!pfd || !s) {
		fprintf(stderr, "Error - Invalid parameters to submission handler.\n");
		exit(EXIT_FAILURE);
	}

	// this is a simple state machine for controlling submissions
	switch (s->state) {
	case SS_NEW:
		// send random nonce
		srand(time(0));
		nonce = htonl((u_int32_t) (RAND_MAX * (rand() / (RAND_MAX + 1.0))));
		if (write(pfd->fd, &(nonce), 4) == -1) {
			fprintf(stderr, "Error - Unable to send session nonce: %s.\n", strerror(errno));
			return(-1);
		}
		s->nonce	= ntohl(nonce);
		s->state	= SS_NONCE_SENT;
		s->bytes_read	= 0;

		break;
	case SS_NONCE_SENT:
		// read secret hash
		if ((rv = session_read_data(pfd->fd, s, (u_char *) s->secret_hash, 128)) == -1) {
			return(-1);
		} else if (rv == 128) {
			s->state = SS_UNAUTHENTICATED;

			if (secret) {
				if ((secret_nonce = calloc(1, strlen(secret) + 4)) == NULL) {
					fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
					return(-1);
				}
				memcpy(secret_nonce, secret, strlen(secret));
				memcpy(&secret_nonce[strlen(secret)], &s->nonce, 4);

				if ((sha512sum = mem_sha512sum((u_char *) secret_nonce, strlen(secret)+4)) == NULL) {
					fprintf(stderr, "Error - Unable to hash secret.\n");
					return(-1);
				}
				free(secret_nonce);
			} else {
				if ((sha512sum = mem_sha512sum((u_char *) &s->nonce, 4)) == NULL) {
					fprintf(stderr, "Error - Unable to hash secret.\n");
					return(-1);
				}
			}

			if (!strncmp(sha512sum, s->secret_hash, 128)) {
				if (verbose > 2) printf("    Valid secret hash read, session authenticated.\n");
				s->state	= SS_AUTHENTICATED;
				s->bytes_read	= 0;
			}

			free(sha512sum);

			if (s->state != SS_AUTHENTICATED) {
				printf("[x] Secret mismatch, dropping session: %d.\n", pfd->fd);
				fflush(stdout);
				return(0);
			}
		} else if (rv == 0) {
			return(0);
		}
		break;
	case SS_AUTHENTICATED:
		// read md5 hash
		if (!s->md5sum && ((s->md5sum = calloc(sizeof(char), 33)) == NULL)) {
			fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		if ((rv = session_read_data(pfd->fd, s, (u_char *) s->md5sum, 32)) == -1) {
			return(-1);
		} else if (rv == 32) {
			s->md5sum[32] = 0;
			if (verbose > 2) printf("    Attack MD5 hash: %s\n", s->md5sum);

			pthread_rwlock_rdlock(&md5sum_trie_lock);
			t = trie_find_memstr(&md5sum_trie, (u_char *) s->md5sum, strlen(s->md5sum));

			if (t != NULL) {
				// md5sum is already in trie
				if (write(pfd->fd, "KNOWN\n", 6) == -1) {
					fprintf(stderr, "Error - Unable to send data request: %s.\n", strerror(errno));
					pthread_rwlock_unlock(&md5sum_trie_lock);
					return(-1);
				}

				// increase instance counter
				((hash*)t->data)->cnt++;
				pthread_rwlock_unlock(&md5sum_trie_lock);
				if (verbose > 1) printf("    MD5 hash is %s (%u instances)\n", ((hash*)t->data)->md5sum, ((hash*)t->data)->cnt);

				if (verbose > 1) printf("[x] Absolute MD5 match found, Rejecting known attack\n");
				fflush(stdout);
				return(0);
			} else {
				// unknown attack hash, send request
				pthread_rwlock_unlock(&md5sum_trie_lock);
				if (write(pfd->fd, "UNKNOWN\n", 8) == -1) {
					fprintf(stderr, "Error - Unable to send data request: %s.\n", strerror(errno));
					return(-1);
				}
			}

			if (verbose > 2) printf("    Unknown attack, submission requested.\n");
			s->state	= SS_REQUEST_SENT;
			s->bytes_read	= 0;
		}
		break;
	case SS_REQUEST_SENT:
		// read protocol
		if ((rv = session_read_data(pfd->fd, s, (u_char *) &(s->protocol), 1)) == -1) {
			return(-1);
		} else if (rv == 1) {
			switch (s->protocol) {
			case 6:
				if (verbose > 2) printf("    Protocol: TCP\n");
				break;
			case 17:
				if (verbose > 2) printf("    Protocol: UDP\n");
				break;
			default:
				if (verbose > 2) printf("    Protocol: unknown\n");
				return(0);
				break;
			}
			s->state	= SS_PROTOCOL_READ;
			s->bytes_read	= 0;
		}
		break;
	case SS_PROTOCOL_READ:
		// read port
		if ((rv = session_read_data(pfd->fd, s, (u_char *) &(s->port), 2)) == -1) {
			return(-1);
		} else if (rv == 2) {
			s->port = ntohs(s->port);
			if (verbose > 2) printf("    Port: %u\n", s->port);
			s->state	= SS_PORT_READ;
			s->bytes_read	= 0;
		}
		break;
	case SS_PORT_READ:
		// read length of attack data
		if ((rv = session_read_data(pfd->fd, s, (u_char *) &(s->attack_len), 4)) == -1) {
			return(-1);
		} else if (rv == 4) {
			s->attack_len = ntohl(s->attack_len);
			if (verbose > 2) printf("    Bytes of attack string: %lu\n", s->attack_len);
			s->state	= SS_ATTACK_LEN_READ;
			s->bytes_read	= 0;
		}
		break;
	case SS_ATTACK_LEN_READ:
		// read length of compressed attack data
		if ((rv = session_read_data(pfd->fd, s, (u_char *) &(s->cattack_len), 4)) == -1) {
			return(-1);
		} else if (rv == 4) {
			s->cattack_len = ntohl(s->cattack_len);
			if (verbose > 2) printf("    Bytes of compressed attack string: %u\n", s->cattack_len);
			s->state	= SS_CATTACK_LEN_READ;
			s->bytes_read	= 0;
		}
		break;
	case SS_CATTACK_LEN_READ:
		// read compressed attack
		if (!s->cattack && ((s->cattack = calloc(1, s->cattack_len)) == NULL)) {
			fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		if ((rv = session_read_data(pfd->fd, s, s->cattack, s->cattack_len)) == -1) {
			return(-1);
		} else if (rv == s->cattack_len) {
			if (verbose > 2) printf("    Attack read.\n");
			s->state	= SS_CATTACK_READ;
			s->bytes_read	= 0;
		} else return(2);
	case SS_CATTACK_READ:
		// read length of HMAC
		s->hmac_len = 0xffff;
		if ((rv = session_read_data(pfd->fd, s, (u_char *) &(s->hmac_len), sizeof(s->hmac_len))) == -1) {
			return(-1);
		} else if (rv == 2) {
			s->hmac_len = ntohs(s->hmac_len);
			if (verbose > 2) printf("    Length of HMAC: %u\n", s->hmac_len);
			s->state	= SS_HMAC_LEN_READ;
			s->bytes_read	= 0;
		}
		break;
	case SS_HMAC_LEN_READ:
		if (!s->hmac && ((s->hmac = calloc(sizeof(u_char), s->hmac_len+1)) == NULL)) {
			fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		if ((rv = session_read_data(pfd->fd, s, (u_char *) s->hmac, s->hmac_len)) == -1) {
			return(-1);
		} else if (rv == s->hmac_len) {
			if (verbose > 2) printf("    HMAC read: %s\n", s->hmac);
			hmac_port = htons(s->port);

			// send OK
			if (write(pfd->fd, "OK\n", 3) == -1) {
				fprintf(stderr, "Error - Unable to send response: %s.\n", strerror(errno));
				return(-1);
			}
			if (verbose > 2) printf("    Response sent.\n");


			// append protocol and port to cattack for HMAC verification
			if ((s->cattack = realloc(s->cattack, s->cattack_len+3)) == NULL) {
				fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
				return(-1);
			}
			memcpy(s->cattack+s->cattack_len, &s->protocol, 1);
			memcpy(s->cattack+s->cattack_len+1, &hmac_port, 2);

			if ((sha512sum = session_hmac(&s->cattack, s->cattack_len+3)) == NULL) {
				fprintf(stderr, "Error - Unable to compute HMAC.\n");
				return(-1);
			}

			// free appended protocol and port memory
			if ((s->cattack = realloc(s->cattack, s->cattack_len)) == NULL) {
				fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
				free(sha512sum);
				return(-1);
			}

			if (strlen(sha512sum) != s->hmac_len || strncmp(sha512sum, s->hmac, s->hmac_len)) {
				printf("[x] HMAC does not match, discarding submission.\n");
				free(sha512sum);
				fflush(stdout);
				return(0);
			}

			free(sha512sum);
			if (verbose > 2) printf("    HMAC is valid.\n");


			// process attack
			if ((s->attack = malloc(sizeof(u_char) * s->attack_len)) == NULL) {
				fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
				return(-1);
			}

			// uncompress attack
			switch (uncompress(s->attack, &(s->attack_len), s->cattack, s->cattack_len)) {
			case Z_OK:
				break;
			case Z_MEM_ERROR:
				fprintf(stderr, "Error - Cannot uncompress attack: Memory error.\n");
				exit(EXIT_FAILURE);
			case Z_BUF_ERROR:
				fprintf(stderr, "Error - Cannot uncompress attack: Buffer error.\n");
				exit(EXIT_FAILURE);
			default:
				fprintf(stderr, "Error - Cannot uncompress attack: Unknown error.\n");
				exit(EXIT_FAILURE);
			}
			if (verbose > 2) printf("    Successfully uncompressed %lu bytes.\n", s->attack_len);

			if (s->protocol == 6)
				printf("[.] Received new attack (%lu bytes from port %u/tcp)\n", s->attack_len, s->port);
			else if (s->protocol == 17)
				printf("[.] Received new attack (%lu bytes from port %u/udp)\n", s->attack_len, s->port);

			// calculate and check md5sum
			md5sum = mem_md5sum(s->attack, s->attack_len);
			if (md5sum && strncmp(md5sum, s->md5sum, 32)) {
				fprintf(stderr, "[x] Attack checksum does not match, discarding submission.\n");
				return(-1);
			}
			free(md5sum);
			free(s->cattack);
			s->cattack = NULL;


			s->bytes_read	= 0;
			s->state	= SS_FINISHED;

			fflush(stdout);
			return(1);
		}
		break;
	default:
		return(-2);
	}

	return(2);
}
