/* sig.c
 *
 * Copyright (C) 2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/types.h>
#include <unistd.h>

#include "cluster.h"
#include "cstr.h"
#include "nebula.h"
#include "session.h"
#include "sha512.h"
#include "sig.h"
#include "stree.h"


#define MAX_SIG_LOGLINE_LEN 120		// break line after 120 chars when logging signatures to stdout

#define INC_CHARCOUNT(a)	printed_chars += a; if (printed_chars >= MAX_SIG_LOGLINE_LEN-1) { printed_chars = 0; printf("\n           "); }

int segment_add(sseg **list, ssize_t *num_frags, sseg seg) {
	ssize_t	i, low, high;

	i = 0;

	// find position in subsequence
	for (i=0, low=0, high=*num_frags; high-low > 0; ) {
		i = (high+low)/2;
		if (seg.min_off > (*list)[i].max_off+(*list)[i].len) high = i;
		else {
			i = (high+low+1)/2;
			low = i;
		}
	}

	// if new segment does not fit (i.e., intersects with a present segment) return an error
	if (i && seg.max_off+seg.len >= (*list)[i-1].min_off) return(-1);

	// insert segment
	if ((*list = realloc(*list, (*num_frags + 1) * sizeof(sseg))) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	memset(&(*list)[*num_frags], 0, sizeof(sseg));
	memmove(&(*list)[i+1], &(*list)[i], (*num_frags-i) * sizeof(sseg));
	memcpy(&(*list)[i], &seg, sizeof(sseg));

	return((*num_frags)++);
}


void build_sig(signature *sig, stree *t, lcatbl *lca_table, stnode **leaves, ssize_t num_leaves, ssize_t min_len, double min_ent) {
	int		i, j, printable, printed_chars;
	double		ent;
	u_int32_t	start;
	ssize_t		sigdata_len, leaf, id, l, r, longnum;
	sseg		*seglist;
	u_char		*sigdata;
	char		*sha512sum;
	trie_node	*n;

	ssize_t		num_frags;
	sseg		seg;

	printable	= 0;
	num_frags	= 0;
	sigdata_len	= 0;
	seglist		= NULL;
	sigdata		= NULL;

	for (i = 0; i < sig->cstr_list.len; i++) {
		start		= sig->cstr_list.elem[i].n->edge_label_start;
		ent		= 0;

		// initialize segment
		memset(&seg, 0, sizeof(sseg));
		seg.len		= sig->cstr_list.elem[i].len;
		seg.min_off	= t->length;
		seg.max_off	= 0;

		if ((min_len && sig->cstr_list.elem[i].len < min_len) || (ent = entropy(t->tree_string+start, sig->cstr_list.elem[i].len)) < min_ent)
			continue;

		// find leaves covered by the current substring
		l = sig->cstr_list.elem[i].l;
		r = sig->cstr_list.elem[i].r;
		while (l > 0 && lca_lookup(lca_table, sig->cstr_list.elem[i].n, leaves[l-1]) == sig->cstr_list.elem[i].n) l--;
		while (r+1 < num_leaves && lca_lookup(lca_table, sig->cstr_list.elem[i].n, leaves[r+1]) == sig->cstr_list.elem[i].n) r++;

		// find minimal and maximal offset for leaves in this substring's subtree
		for (leaf = l; leaf < num_leaves && leaf <= r; leaf++) {
			id	= leaves[leaf]->string_id-1;
			start	= leaves[leaf]->path_position;

			if (id == 0) {
				seg.org_off = start - t->string_offset[id]-1;
			}
			if (start - t->string_offset[id]-1 < seg.min_off) seg.min_off = start - t->string_offset[id]-1;
			if (start - t->string_offset[id]-1 > seg.max_off) seg.max_off = start - t->string_offset[id]-1;
		}

		// add segment to common subsequence
		segment_add(&seglist, &num_frags, seg);
	}


	// store signature info into a contiguous memory area and calculate hash
	for (i=num_frags; i; i--) {
		// append next signature segment
		if ((sigdata = realloc(sigdata, sigdata_len + (2*sizeof(ssize_t)) + seglist[i-1].len)) == NULL) {
			fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
			free(seglist);
			free(sigdata);
			return;
		}
		memset(&sigdata[sigdata_len], 0, (2*sizeof(ssize_t)) + seglist[i-1].len);

		// copy offset info
		if (i == num_frags) {
			longnum = seglist[num_frags-1].max_off+seglist[num_frags-1].len;
		} else {
			longnum =  seglist[i-1].min_off - seglist[i].max_off - seglist[i].len;
		}
		memcpy(&sigdata[sigdata_len], &longnum, sizeof(ssize_t));
		sigdata_len += sizeof(ssize_t);

		// copy depth info
		if (i == num_frags) {
			longnum = seglist[num_frags-1].min_off > seglist[num_frags-1].len ? seglist[num_frags-1].min_off - seglist[num_frags-1].len : 0;
		} else {
			longnum = seglist[i-1].max_off + seglist[i-1].len - seglist[i].min_off;
		}
		memcpy(&sigdata[sigdata_len], &longnum, sizeof(ssize_t));
		sigdata_len += sizeof(ssize_t);

		// copy segment bytes
		for (j=0; j<seglist[i-1].len; j++) sigdata[sigdata_len+j] = t->tree_string[seglist[i-1].org_off+1+j];

		sigdata_len += seglist[i-1].len;
	}

	// hash signature and do a trie lookup
	sha512sum = mem_sha512sum(sigdata, sigdata_len);
	printf("[=] signature hash: %s\n", sha512sum);

	pthread_rwlock_rdlock(&sighash_trie_lock);
	n = trie_find_memstr(&sighash_trie, (u_char *) sha512sum, 128);
	pthread_rwlock_unlock(&sighash_trie_lock);

	if (n != NULL) {
		// signature was generated earlier so skip it
		printf("[=] signature was already published.\n");

		// increase size threshold for this cluster
		sig->cl->threshold *= 1.3;

		free(sigdata);
		free(sha512sum);
		free(seglist);
		return;
	} else {
		// insert signature hash into trie
		pthread_rwlock_wrlock(&sighash_trie_lock);
		n = trie_memins(&sighash_trie, (u_char *) sha512sum, 128, NULL);
		n->data = sig->cl;
		pthread_rwlock_unlock(&sighash_trie_lock);
	}
	free(sigdata);
	free(sha512sum);
	

	switch (sig->type) {
	case SIG_RAW:
		// print raw signature
		printf("\n---- Signature type: raw ------------------------------------------------------------------------------\n");
		for (i=num_frags; i; i--) {
			printf("#\033[1;32m[%lu]\033[1;31m[\033[1;0m", (long unsigned int) seglist[i-1].min_off);
			for (j=1, printed_chars=0; j<=seglist[i-1].len; j++) {
				putchar(isprint(t->tree_string[seglist[i-1].org_off+j]) ? t->tree_string[seglist[i-1].org_off+j] : '.');
				INC_CHARCOUNT(1);
			}
			printf("\033[1;31m]\033[1;32m[%lu]\033[1;0m  ", (long unsigned int) seglist[i-1].max_off+seglist[i-1].len);
		}
		printf("-------------------------------------------------------------------------------------------------------\n\n");
		break;
	case SIG_SNORT:
		if (!sig->cl->sig_id) {
			// increase global sid
			pthread_rwlock_wrlock(&sidlock);
			sig->cl->sig_id = ++global_sid;
			pthread_rwlock_unlock(&sidlock);
		}

		// increase revision number
		sig->cl->sig_rev++;

		if (rules_file) {
			// append snort signature to rules file
			if (append_snortsig_to_rulefile(rules_file, t, sig, seglist, num_frags) > 0) {
				if (snort_pid) {
					printf("[#] forcing snort to reload rule set\n");
					// make snort reload its configuration
					if (kill(snort_pid, SIGHUP) != 0) {
						fprintf(stderr, "Error - Sending SIGHUP to snort (%u) failed: %s.\n", snort_pid, strerror(errno));
					}
				}
			}
		}

		// print snort signature
		printf("\n---- Signature type: snort ----------------------------------------------------------------------------\n");

		// protocol info
		if (sig->proto_type == SIGPROTO_UDP) {
			printf("alert udp ");
		} else {
			printf("alert tcp ");
		}

		// port info
		if (sig->port_type == SIGPORT_UNIQ && sig->port) {
			printf("any any -> $HOME_NET %u (msg: \"nebula rule %u rev. %u\";", sig->port, sig->cl->sig_id, sig->cl->sig_rev); 
		} else {
			printf("any any -> $HOME_NET any (msg: \"nebula rule %u rev. %u\";", sig->cl->sig_id, sig->cl->sig_rev); 
		}

		if (num_frags) {
			printf(" \\\n content: \"");
			printable = 1;
			for (j=1,printed_chars=0; j<=seglist[num_frags-1].len; j++) {
				if (!isprint(t->tree_string[seglist[num_frags-1].org_off+j])
				  || t->tree_string[seglist[num_frags-1].org_off+j] == '|'
				  || t->tree_string[seglist[num_frags-1].org_off+j] == '$') {
					if (printable) {
						printable = 0;
						putchar('|');
						INC_CHARCOUNT(1);
					} else if (printed_chars != 0) {
						putchar(' ');
						INC_CHARCOUNT(1);
					}
					printf("%02x", t->tree_string[seglist[num_frags-1].org_off+j]);
					INC_CHARCOUNT(2);
				} else if (isprint(t->tree_string[seglist[num_frags-1].org_off+j])) {
					if (!printable) {
						printable = 1;
						putchar('|');
						INC_CHARCOUNT(1);
					}
					// quote special characters
					switch (t->tree_string[seglist[num_frags-1].org_off+j]) {
					case ':':
					case ';':
					case '\\':
					case '"':
						printf("\\%c", t->tree_string[seglist[num_frags-1].org_off+j]);
						INC_CHARCOUNT(2);
						break;
					default:
						putchar(t->tree_string[seglist[num_frags-1].org_off+j]);
						INC_CHARCOUNT(1);
					}
				}
			}
			if (!printable) putchar('|');
			INC_CHARCOUNT(1);
			printf("\"; offset: %lu; depth: %lu;",
				(long unsigned int) (seglist[num_frags-1].min_off > seglist[num_frags-1].len ?
					seglist[num_frags-1].min_off - seglist[num_frags-1].len : 0),
				(long unsigned int) seglist[num_frags-1].max_off+seglist[num_frags-1].len);

			// process other segments
			for (i=num_frags-1; i; i--) {
				printable = 1;
				printf(" \\\n content: \"");
				for (j=1, printed_chars=0; j<=seglist[i-1].len; j++) {
					if (!isprint(t->tree_string[seglist[i-1].org_off+j])
					  || t->tree_string[seglist[i-1].org_off+j] == '|'
					  || t->tree_string[seglist[i-1].org_off+j] == '$') {
						if (printable) {
							printable = 0;
							putchar('|');
							INC_CHARCOUNT(1);
						} else if (printed_chars != 0) {
							putchar(' ');
							INC_CHARCOUNT(1);
						}
						printf("%02x", t->tree_string[seglist[i-1].org_off+j]);
						INC_CHARCOUNT(2);
					} else if (isprint(t->tree_string[seglist[i-1].org_off+j])) {
						if (!printable) {
							printable = 1;
							putchar('|');
							INC_CHARCOUNT(1);
						}
						// quote special characters
						switch (t->tree_string[seglist[i-1].org_off+j]) {
						case ':':
						case ';':
						case '\\':
						case '"':
						case '|':
							printf("\\%c", t->tree_string[seglist[i-1].org_off+j]);
							INC_CHARCOUNT(2);
							break;
						default:
							putchar(t->tree_string[seglist[i-1].org_off+j]);
							INC_CHARCOUNT(1);
						}
					}
				}
				if (!printable) putchar('|');
				INC_CHARCOUNT(1);

				printf("\"; distance: %lu; within: %lu;",
					(long unsigned int) seglist[i-1].min_off - seglist[i].max_off - seglist[i].len,
					(long unsigned int) seglist[i-1].max_off + seglist[i-1].len - seglist[i].min_off);
			}
			printf(" \\\n sid: %u; rev: %u;)\n", sig->cl->sig_id, sig->cl->sig_rev);
		}
		printf("-------------------------------------------------------------------------------------------------------\n\n");
		break;
	default:
		fprintf(stderr, "Error - Signature type %i is currently not supported.\n", sig->type);
		exit(EXIT_FAILURE);
	}

	free(seglist);

	fflush(stdout);
	return;
}


void *pt_pollgq(void *param) {
	qelem			*elem;

	for (;;) {
		if (repsem_wait(&siggen_semaphore) == -1) {
			fprintf(stderr, "Error - Unable to decrement semaphore: %s.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		pthread_mutex_lock(&siggenq_mutex);
		elem = queue_cuthead(siggenq);
		pthread_mutex_unlock(&siggenq_mutex);

		if (elem) {
			gensig(elem->data);
			free(elem);
		}
	}

	return((void *) 1);
}


void *gensig(void *cl) {
	cluster		*attacks = cl;
	lchar		*content;
	stnode		*lca, **leaves;
	u_int32_t	i, acnt, node_id, num_leaves, strllen;
	lchar		*strlist;
	u_char		*byte;
	qelem		*qe;
	submission	*attack;
	stree		*gst;
	lcatbl		*lca_table;
	signature	sig;

	lca		= NULL;
	leaves		= NULL;
	strlist		= NULL;
	strllen		= 0;

	gst		= NULL;
	content		= NULL;

	acnt		= 0;

	memset(&sig, 0, sizeof(signature));
	sig.cl		= cl;
	sig.type	= SIG_SNORT;
	sig.port_type	= SIGPORT_NOTSET;
	sig.proto_type	= SIGPROTO_NONE;

	// create suffix tree

	if ((gst = ST_CreateTree()) == NULL) {
		fprintf(stderr, "Error - Unable initialize generalized suffix tree.\n");
		exit(EXIT_FAILURE);
	}

	CL_RLOCK(attacks);
	for (qe = attacks->hq->head; qe; qe = qe->next) {
		attack = ((hash *)qe->data)->submission;

		if (!attack) continue;

		if (strllen > 0xffffffff - attack->attack_len) {
			fprintf(stderr, "Warning - String list would exceed maximum length, skipping attack.\n");
			continue;
		}
		acnt++;


		// concatenate attack strings
		// byte strlist[0] must not be used, it simplifies the GST algorithm which uses offset 0 for the root node
		if ((strlist = realloc(strlist, sizeof(lchar) * (strllen + attack->attack_len + 2))) == NULL) {
			fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		// extend string offset array
		if ((gst->string_offset = realloc(gst->string_offset, acnt * sizeof(u_int32_t))) == NULL) {
			fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		gst->string_offset[acnt-1] = strllen;

		for (i = 0; i < attack->attack_len; i++) {
			strlist[strllen+i+1] = attack->attack[i];
		}
		strllen += attack->attack_len + 1;
		strlist[strllen] = acnt << 8;

		// process port info
		if (sig.port == 0) {
			sig.port	= attack->port;
			sig.port_type	= SIGPORT_UNIQ;
		} else if (sig.port != attack->port) sig.port_type = SIGPORT_ANY;

		// process protocol info
		switch (attack->protocol) {
		case 6:
			sig.proto_type &= SIGPROTO_TCP;
			break;
		case 17:
			sig.proto_type &= SIGPROTO_UDP;
			break;
		default:
			sig.proto_type = SIGPROTO_INVALID;
		}
	}
	CL_ULOCK(attacks);

	// print concatenated string
	if (verbose > 2) {
		printf("Concatenated string (%u) is '", strllen);
		for (i=1; i <= strllen; i++) {
			byte = (u_char *) &strlist[i];
			if (strlist[i] && !(strlist[i] & 0x000000ff))
				printf("\033[1;31m[%u]\033[1;0m", strlist[i] >> 8);
			else {
				putchar(isprint(*byte) ? *byte : '.');
			}
		}
		printf("'\n\n");
	}


	// construct generalized suffix tree
	if (verbose) printf("    Constructing generalized suffix tree.\n");
	pthread_mutex_lock(&siggen_mutex);
	if ((gst = ST_AddStringList(gst, strlist, strllen)) == NULL) {
		fprintf(stderr, "Error - Unable to add attack to suffix tree.\n");
		free(strlist);
		exit(EXIT_FAILURE);
	}
	pthread_mutex_unlock(&siggen_mutex);


	// compute LCA table for tree nodes
	node_id		= 0;
	num_leaves	= 0;
	if ((leaves = calloc(1, sizeof(stnode *))) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	set_ids(gst, gst->root, &leaves, &num_leaves, &node_id);

	if (verbose) printf("    Computing LCA table.\n");
	if ((lca_table = lca_prep(gst)) == NULL) {
		fprintf(stderr, "Error - Unable to compute LCA table.\n");
		exit(EXIT_FAILURE);
	}


	// calculate common subtree
	if (verbose) printf("    Calculating common subtree.\n");

	// process leaves from left to right by sequentially moving left and right interval boundaries
	memset(&sig.cstr_list, 0, sizeof(substr_list));
	if (calc_common_subtree(gst, &sig.cstr_list, min_sstr_len, lca_table, leaves, num_leaves, acnt) == -1) {
		fprintf(stderr, "Error - Unable to find substrings.\n");
		exit(EXIT_FAILURE);
	}
	if (verbose) printf("    Generalized suffix tree contains a common subtree with %lu leaves.\n", (long unsigned int) sig.cstr_list.len);
	if (verbose > 3) ST_PrintTree(gst, 0);


	// Extract common substrings
	if (verbose) printf("    Sorting common substrings by length.\n");
	qsort(sig.cstr_list.elem, sig.cstr_list.len, sizeof(substr), substr_len_cmp);

	if (verbose) printf("    Extracting common substrings with length >= %lu.\n", (long unsigned int) min_sstr_len);
	if (verbose > 2) for (i = 0; i < sig.cstr_list.len && sig.cstr_list.elem[i].len >= min_sstr_len; i++)
		list_substrings(gst, sig.cstr_list.elem[i].n, sig.cstr_list.elem[i].len);

	build_sig(&sig, gst, lca_table, leaves, num_leaves, min_sstr_len, min_sstr_ent);


	// free data structures
	free(sig.cstr_list.elem);

	ST_DeleteTree(gst);

	lca_free(lca_table);
	free(leaves);

	if (verbose) printf("    Done.\n");
	return(0);
}


int append_snortsig_to_rulefile(const char *filename, const stree *t, const signature *sig, const sseg *seglist, const ssize_t num_frags) {
	int	i, j, printable;
	FILE	*rfile;

	// lock mutex to make sure only one thread writes at the same time
	pthread_mutex_lock(&sigwrite_mutex);

	if ((rfile = fopen(filename, "a")) == NULL) {
		fprintf(stderr, "Error - Unable to open rules file: %s.\n", strerror(errno));
		return(-1);
	}

	fprintf(rfile, "#-----------------------------------------------------------------------------------------------------------------------\n");

	// protocol info
	if (sig->proto_type == SIGPROTO_UDP) {
		fprintf(rfile, "alert udp ");
	} else {
		fprintf(rfile, "alert tcp ");
	}

	// port info
	if (sig->port_type == SIGPORT_UNIQ && sig->port) {
		fprintf(rfile, "any any -> $HOME_NET %u (msg: \"nebula rule %u rev. %u\";", sig->port, sig->cl->sig_id, sig->cl->sig_rev); 
	} else {
		fprintf(rfile, "any any -> $HOME_NET any (msg: \"nebula rule %u rev. %u\";", sig->cl->sig_id, sig->cl->sig_rev); 
	}

	if (num_frags) {
		fprintf(rfile, " \\\n content: \"");
		printable = 1;
		for (j=1; j<=seglist[num_frags-1].len; j++) {
			if (!isprint(t->tree_string[seglist[num_frags-1].org_off+j])
			  || t->tree_string[seglist[num_frags-1].org_off+j] == '|'
			  || t->tree_string[seglist[num_frags-1].org_off+j] == '$') {
				if (printable) {
					printable = 0;
					fprintf(rfile, "|");
				} else fprintf(rfile, " ");
				fprintf(rfile, "%02x", t->tree_string[seglist[num_frags-1].org_off+j]);
			} else if (isprint(t->tree_string[seglist[num_frags-1].org_off+j])) {
				if (!printable) {
					printable = 1;
					fprintf(rfile, "|");
				}
				// quote special characters
				switch (t->tree_string[seglist[num_frags-1].org_off+j]) {
				case ':':
				case ';':
				case '\\':
				case '"':
					fprintf(rfile, "\\%c", t->tree_string[seglist[num_frags-1].org_off+j]);
					break;
				default:
					fprintf(rfile, "%c", t->tree_string[seglist[num_frags-1].org_off+j]);
				}
			}
		}
		if (!printable) fprintf(rfile, "|");
		fprintf(rfile, "\"; offset: %lu; depth: %lu;",
			(long unsigned int) (seglist[num_frags-1].min_off > seglist[num_frags-1].len ? seglist[num_frags-1].min_off - seglist[num_frags-1].len : 0),
			(long unsigned int) seglist[num_frags-1].max_off+seglist[num_frags-1].len);

		// process other segments
		for (i=num_frags-1; i; i--) {
			printable = 1;
			fprintf(rfile, " \\\n content: \"");
			for (j=1; j<=seglist[i-1].len; j++) {
				if (!isprint(t->tree_string[seglist[i-1].org_off+j])
				  || t->tree_string[seglist[i-1].org_off+j] == '|'
				  || t->tree_string[seglist[i-1].org_off+j] == '$') {
					if (printable) {
						printable = 0;
						fprintf(rfile, "|");
					} else fprintf(rfile, " ");
					fprintf(rfile, "%02x", t->tree_string[seglist[i-1].org_off+j]);
				} else if (isprint(t->tree_string[seglist[i-1].org_off+j])) {
					if (!printable) {
						printable = 1;
						fprintf(rfile, "|");
					}
					// quote special characters
					switch (t->tree_string[seglist[i-1].org_off+j]) {
					case ':':
					case ';':
					case '\\':
					case '"':
					case '|':
						fprintf(rfile, "\\%c", t->tree_string[seglist[i-1].org_off+j]);
						break;
					default:
						fprintf(rfile, "%c", t->tree_string[seglist[i-1].org_off+j]);
					}
				}
			}
			if (!printable) fprintf(rfile, "|");

			fprintf(rfile, "\"; distance: %lu; within: %lu;",
				(long unsigned int) seglist[i-1].min_off - seglist[i].max_off - seglist[i].len,
				(long unsigned int) seglist[i-1].max_off + seglist[i-1].len - seglist[i].min_off);
		}
		fprintf(rfile, " \\\n sid: %u; rev: %u;)\n", sig->cl->sig_id, sig->cl->sig_rev);
	}

	fclose(rfile);

	// unlock mutex
	pthread_mutex_unlock(&sigwrite_mutex);

	return(1);
}
