/* signals.c
 *
 * Copyright (C) 2007-2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "nebula.h"
#include "signals.h"
#include "util.h"


#ifdef SIGRTMIN
					// SIGRTMIN+[0..2] are used for thread control (not available on all POSIX platforms)
#  define SIGINCV	(SIGRTMIN+3)	// signal to increase verbosity level
#  define SIGDECV	(SIGRTMIN+4)	// signal to decrease verbosity level
#  define SIGINCC	(SIGRTMIN+5)	// signal to increase similarity threshold by 5
#  define SIGDECC	(SIGRTMIN+6)	// signal to decrease similarity threshold by 5
#endif


void handle_ctrl_signal(int sig) {
	fflush(stdout);
#ifdef SIGINCV
	if (sig == SIGINCV) {
		verbose++;
		if (write(STDOUT_FILENO, "[*] Verbosity level increased.\n", 31) == -1) {}
		return;
	}
#endif
#ifdef SIGDECV
	if (sig == SIGDECV) {
		if (!verbose) return;
		verbose--;
		if (write(STDOUT_FILENO, "[*] Verbosity level decreased.\n", 31) == -1) {}
		return;
	}
#endif
#ifdef SIGINCC
	if (sig == SIGINCC) {
		if (sim_threshold == 100.0) return;
		if (sim_threshold < 95.0) {
			sim_threshold += 5.0;
			if (write(STDOUT_FILENO, "[*] Cluster criterion increased by 5 percent\n", 45) == -1) {}
		} else {
			sim_threshold = 100.0;
			if (write(STDOUT_FILENO, "[*] Cluster criterion increased to 100 percent\n", 46) == -1) {}
		}
		return;
	}
#endif
#ifdef SIGDECC
	if (sig == SIGDECC) {
		if (sim_threshold == 0.0) return;
		if (sim_threshold > 5.0) {
			sim_threshold -= 5.0;
			if (write(STDOUT_FILENO, "[*] Cluster criterion decreased by 5 percent\n", 45) == -1) {}
		} else {
			sim_threshold = 0.0;
			if (write(STDOUT_FILENO, "[*] Cluster criterion decreased to 0 percent\n", 45) == 1) {}
		}
		return;
	}
#endif
	return;
}

void handle_term_signal(int sig) {
	if (sig == SIGINT || sig == SIGTERM) {
		// free memory and quit
		fflush(stdout);
		if (verbose) if (write(STDOUT_FILENO, "[*] Termination forced, cleaning up.\n", 37) == -1) {}
		cleanup();
		if (write(STDOUT_FILENO, "[*] Done.\n", 10) == -1) {}
		exit(EXIT_SUCCESS);
	}

	// signal indicates an error
	if (write(STDOUT_FILENO, "\n[*] Premature termination forced.\n", 35) == -1) {}
	exit(EXIT_FAILURE);
}


void set_signal_handlers(void) {
	u_char			i;
	struct sigaction	s_action;
	static int		termsigs[] = {
#ifdef HAVE_SIGBUS
		SIGBUS,
#endif
		SIGHUP,
		SIGILL,
		SIGINT,
		SIGQUIT,
		SIGSEGV,
		SIGTERM
	};


#ifdef SIGRTMIN
	// install handler for control realtime signals
	memset(&s_action, 0, sizeof(struct sigaction));
	s_action.sa_handler	= handle_ctrl_signal;
#ifdef SA_RESTART
	s_action.sa_flags	|= SA_RESTART;
#endif
	for (i = 3; i < 8; i++) {
		if (sigaction(SIGRTMIN+i, &s_action, NULL) == -1) {
			fprintf(stdout, "  Error - Unable to install handler for signal %d: %s.\n", SIGRTMIN+i, strerror(errno));
			exit(EXIT_FAILURE);
		}
	}
#endif


	// install handler for signals which terminate the process 
	memset(&s_action, 0, sizeof(struct sigaction));
	s_action.sa_handler	= handle_term_signal;
#ifdef SA_RESTART
	s_action.sa_flags	|= SA_RESTART;
#endif
	for (i = 0; i < sizeof(termsigs)/sizeof(termsigs[0]); i++) {
		if (sigaction(termsigs[i], &s_action, NULL) == -1) {
			fprintf(stdout, "  Error - Unable to install handler for signal %d: %s.\n", termsigs[i], strerror(errno));
			exit(EXIT_FAILURE);
		}
	}

	return;
}
