/* stree.c
 *
 * Copyright (C) 2007-2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Some of this code is based on Shlomo Yona's suffix tree implementation.
 */

#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include <unistd.h>

#include "avl.h"
#include "cstr.h"
#include "lca.h"
#include "nebula.h"
#include "stree.h"
#include "util.h"


typedef enum SKIP_TYPE { skip, no_skip } SKIP_TYPE;		/* Used in function trace_string for skipping (Ukkonen's Skip Trick). */
typedef enum RULE_2_TYPE { new_son, split } RULE_2_TYPE;	/* Used in function apply_rule_2 - two types of rule 2 - see function for more details.*/
typedef enum LAST_POS_TYPE { last_char_in_edge, other_char } LAST_POS_TYPE; /* Signals whether last matching position is the last one of the current edge */
stnode *suffixless;		/* Used to mark the node that has no suffix link yet. By Ukkonen, it will have one by the end of the current phase. */

typedef struct {
	u_int32_t begin;
	u_int32_t end;
} PATH;

typedef struct {
	stnode *node;
	u_int32_t edge_pos;
} POS;


// set integers as node IDs while traversing a tree in inorder
void set_ids(stree *t, stnode *n, stnode ***leaves, u_int32_t *num_leaves, u_int32_t *id) {
	u_int32_t	start, end;
	stnode		*tmpnode;

	if (!n) return;

	n->id = ++(*id);
	if (!n->sons) {
		// set string id
		start	= n->edge_label_start;
		end	= get_node_label_end(t, n);
		if (start) for (; start <= end; start++) {
			if (t->tree_string[start] && !(t->tree_string[start] & 0x000000ff)) {
				if ((*leaves = realloc(*leaves, sizeof(stnode *) * ((*num_leaves)+1))) == NULL) {
					fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
					exit(EXIT_FAILURE);
				}
				(*leaves)[*num_leaves] = n;
				(*num_leaves)++;

				n->string_id = t->tree_string[start] >> 8;
				break;
			}
		}
	}
	for (tmpnode = n->sons; tmpnode; tmpnode = tmpnode->right_sibling) set_ids(t, tmpnode, leaves, num_leaves, id);
	return;
}


stnode * create_node(stree *tree, stnode *father, u_int32_t start, u_int32_t end, u_int32_t position) {
	stnode	*node;

	if ((node = calloc(1, sizeof(stnode))) == NULL) {
		fprintf(stderr, "Error - Unable to create tree node: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	tree->size++;

	/* Initialize node fields. For detailed description of the fields see suffix_tree.h */
	node->father		= father;
	node->path_position	= position;
	node->edge_label_start	= start;
	node->edge_label_end	= end;
	node->id		= tree->size;

	return(node);
}


inline stnode * find_son(stree * tree, stnode * node, lchar sym) {
	for (node = node->sons; node != 0 && tree->tree_string[node->edge_label_start] != sym; node = node->right_sibling);
	return(node);
}


inline u_int32_t get_node_label_end(stree *tree, stnode* node) {
	if (node->sons == 0) return tree->e;	/* If it's a leaf - return e */
	return(node->edge_label_end);		/* If it's not a leaf - return its real end */
}


inline u_int32_t get_node_label_length(stree *tree, stnode* node) {
  return(get_node_label_end(tree, node)-node->edge_label_start+1);
}


inline char is_last_char_in_edge (stree *tree, stnode *node, u_int32_t edge_pos) {
  return(edge_pos == get_node_label_length(tree, node) - 1 ? 1 : 0);
}


inline void connect_siblings (stnode * left_sib, stnode * right_sib) {
	if (left_sib)	left_sib->right_sibling = right_sib;
	if (right_sib)	right_sib->left_sibling = left_sib;
}


stnode *apply_extension_rule_2(stree *tree, stnode *node, u_int32_t edge_label_begin, u_int32_t edge_label_end,
	u_int32_t path_pos, u_int32_t edge_pos, RULE_2_TYPE type) {
	stnode *new_leaf, *new_internal, *son;

	/*-------new_son-------*/
	if (type == new_son) {
		/* Create a new leaf (4) with the characters of the extension */
		new_leaf = create_node(tree, node, edge_label_begin, edge_label_end, path_pos);

		/* Connect new_leaf (4) as the new son of node (1) */
		son = node->sons;
		while (son->right_sibling) son = son->right_sibling;
		connect_siblings(son, new_leaf);

		return(new_leaf);	/* return (4) */
	}
	/*-------split-------*/
	/* Create a new internal node (3) at the split point */
	new_internal = create_node(tree, node->father, node->edge_label_start, node->edge_label_start + edge_pos, node->path_position);

	/* Update the node (1) incoming edge starting index (it now starts where node (3) incoming edge ends) */
	node->edge_label_start += edge_pos+1;

	/* Create a new leaf (2) with the characters of the extension */
	new_leaf = create_node(tree, new_internal, edge_label_begin, edge_label_end, path_pos);

	/* Connect new_internal (3) where node (1) was, connect (3) with (1)'s left sibling, connect (3) with (1)'s right sibling */
	connect_siblings(node->left_sibling, new_internal);
	connect_siblings(new_internal, node->right_sibling);
	node->left_sibling = 0;

	/* Connect (3) with (1)'s father */
	if (new_internal->father->sons == node) new_internal->father->sons = new_internal;

	/* Connect new_leaf (2) and node (1) as sons of new_internal (3) */
	new_internal->sons = node;
	node->father = new_internal;
	connect_siblings(node, new_leaf);

	/* return (3) */
	return(new_internal);
}


stnode * trace_single_edge(stree *tree, stnode *node, PATH str, u_int32_t *edge_pos, u_int32_t *chars_found, SKIP_TYPE type, int *search_done) {
	stnode		*cont_node;
	u_int32_t	length, str_len;

	/* Set default return values */
	*search_done	= 1;
	*edge_pos	= 0;

	/* Search for the first character of the string in the outcoming edge of node */
	if ((cont_node = find_son(tree, node, tree->tree_string[str.begin])) == NULL) {
		/* Search is done, string not found */
		*edge_pos	= get_node_label_length (tree, node) - 1;
		*chars_found	= 0;
		return(node);
	}

	/* Found first character - prepare for continuing the search */
	node	= cont_node;
	length	= get_node_label_length (tree, node);
	str_len	= str.end - str.begin + 1;

	/* Compare edge length and string length. */
	/* If edge is shorter then the string being searched and skipping is enabled - skip edge */
	if (type == skip) {
		if (length <= str_len) {
			(*chars_found)	= length;
			(*edge_pos)	= length - 1;
			if (length < str_len) *search_done = 0;
		} else {
			(*chars_found)	= str_len;
			(*edge_pos)	= str_len - 1;
		}
		return node;
	} else {
		/* Find minimum out of edge length and string length, and scan it */
		if (str_len < length) length = str_len;

		for (*edge_pos = 1, *chars_found = 1; *edge_pos < length; (*chars_found)++, (*edge_pos)++) {
			/* Compare current characters of the string and the edge. If equal - continue */
			if (tree->tree_string[node->edge_label_start + *edge_pos] != tree->tree_string[str.begin + *edge_pos]) {
				(*edge_pos)--;
				return(node);
			}
		}
	}

	(*edge_pos)--;						/* The loop has advanced *edge_pos one too much */

	if ((*chars_found) < str_len) *search_done = 0;		/* Search is not done yet */

	return(node);
}


stnode * trace_string(stree * tree, stnode *node, PATH str, u_int32_t *edge_pos, u_int32_t *chars_found, SKIP_TYPE type) {
	u_int32_t	edge_chars_found;
	int		search_done = 0;

	*chars_found = 0;
	while (!search_done) {
		*edge_pos		= 0;
		edge_chars_found	= 0;
		node			= trace_single_edge (tree, node, str, edge_pos, &edge_chars_found, type, &search_done);
		str.begin		+= edge_chars_found;
		*chars_found		+= edge_chars_found;
	}
	return node;
}


u_int32_t ST_FindSubstring(stree *tree, lchar *W, u_int32_t P) {
	stnode		*node	= find_son(tree, tree->root, W[0]);
	u_int32_t	k, j	= 0, node_label_end;

	/* Scan nodes down from the root untill a leaf is reached or the substring is found */
	while (node != 0) {
		k		= node->edge_label_start;
		node_label_end	= get_node_label_end (tree, node);

		/* Scan a single edge - compare each character with the searched string */
		for (; j < P && k <= node_label_end && tree->tree_string[k] == W[j]; j++, k++);

		/* Checking which of the stopping conditions are true */
		if (j == P) return node->path_position;					/* W was found, it is a substring. Return its path starting index */
		else if (k > node_label_end) node = find_son (tree, node, W[j]);	/* Current edge is found to match, continue to next edge */
		else return ST_ERROR;							/* One non-matching symbols is found - W is not a substring */
	}
	return ST_ERROR;
}


void follow_suffix_link(stree *tree, POS *pos) {
	PATH		gama;			/* gama is the string between node and its father, in case node doesn't have a suffix link */
	u_int32_t	chars_found = 0;	/* dummy argument for trace_string function */

	if (pos->node == tree->root) return;

	/* If node has no suffix link yet or in the middle of an edge - remember the
	   edge between the node and its father (gama) and follow its father's suffix
	   link (it must have one by Ukkonen's lemma). After following, trace down 
	   gama - it must exist in the tree (and thus can use the skip trick - see 
	   trace_string function description) */
	if (!pos->node->suffix_link || !is_last_char_in_edge(tree, pos->node, pos->edge_pos)) {
		/* If the node's father is the root, than no use following it's link (it 
		   is linked to itself). Tracing from the root (like in the naive 
		   algorithm) is required and is done by the calling function SEA uppon 
		   recieving a return value of tree->root from this function */
		if (pos->node->father == tree->root) {
			pos->node = tree->root;
			return;
		}

		gama.begin	= pos->node->edge_label_start;						/* Store gama - the indices of incoming edge */
		gama.end	= pos->node->edge_label_start + pos->edge_pos;

		pos->node	= pos->node->father->suffix_link;						/* Follow father's suffix link */
		pos->node	= trace_string(tree, pos->node, gama, &(pos->edge_pos), &chars_found, skip);	/* Down-walk gama back to slink's son */
	} else {
		/* If a suffix link exists - just follow it */
		pos->node	= pos->node->suffix_link;
		pos->edge_pos	= get_node_label_length (tree, pos->node)-1;
	}
}



void SEA(stree *tree, POS *pos, PATH str, u_int32_t *rule_applied, char after_rule_3) {
	u_int32_t	chars_found	= 0,
			path_pos	= str.begin;
	stnode		*tmp;


	/* Follow suffix link only if it's not the first extension after rule 3 was applied */
	if (!after_rule_3) follow_suffix_link(tree, pos);

	/* If node is root - trace whole string starting from the root, else - trace last character only */
	if (pos->node == tree->root)
		pos->node = trace_string (tree, tree->root, str, &(pos->edge_pos), &chars_found, no_skip);
	else {
		str.begin	= str.end;
		chars_found	= 0;

		/* Consider 2 cases: */
		if (is_last_char_in_edge (tree, pos->node, pos->edge_pos)) {
			/* 1. last character matched is the last of its edge */
			/* Trace only last symbol of str, search in the  NEXT edge (node) */
			if ((tmp = find_son (tree, pos->node, tree->tree_string[str.end])) != NULL) {
				pos->node	= tmp;
				pos->edge_pos	= 0;
				chars_found	= 1;
			}
		} else {
			/* 2. last character matched is NOT the last of its edge */
			/* Trace only last symbol of str, search in the CURRENT edge (node) */
			if (tree->tree_string[pos->node->edge_label_start + pos->edge_pos + 1] == tree->tree_string[str.end]) {
				pos->edge_pos++;
				chars_found 	= 1;
			}
		}
	}

	/* If whole string was found - rule 3 applies */
	if (chars_found == str.end-str.begin+1) {
		*rule_applied = 3;
		/* If there is an internal node that has no suffix link yet (only one may 
		   exist) - create a suffix link from it to the father-node of the 
		   current position in the tree (pos) */
		if (suffixless) {
			suffixless->suffix_link = pos->node->father;
			suffixless = NULL;
		}
		return;
	}

	/* If last char found is the last char of an edge - add a character at the next edge */
	if (is_last_char_in_edge(tree, pos->node, pos->edge_pos) || pos->node == tree->root) {
		if (pos->node->sons) {		/* Decide whether to apply rule 2 (new_son) or rule 1 */
			/* Apply extension rule 2 new son - a new leaf is created and returned by apply_extension_rule_2 */
			apply_extension_rule_2(tree, pos->node, str.begin+chars_found, str.end, path_pos, 0, new_son);
			*rule_applied = 2;

			/* If there is an internal node that has no suffix link yet (only one 
			   may exist) - create a suffix link from it to the father-node of the 
			   current position in the tree (pos) */
			if (suffixless) {
				suffixless->suffix_link = pos->node;
				suffixless = NULL;
			}
		}

	} else {
		/* Apply extension rule 2 split - a new node is created and returned by apply_extension_rule_2 */
		tmp = apply_extension_rule_2(tree, pos->node, str.begin + chars_found, str.end, path_pos, pos->edge_pos, split);

		if (suffixless) suffixless->suffix_link = tmp;
		/* Link root's sons with a single character to the root */
		if (get_node_label_length(tree, tmp) == 1 && tmp->father == tree->root) {
			tmp->suffix_link	= tree->root;
			suffixless		= 0;		/* Marks that no internal node with no suffix link exists */
		} else suffixless = tmp;			/* Mark tmp as waiting for a link */

		/* Prepare pos for the next extension */
		pos->node	= tmp;
		*rule_applied	= 2;
	}
}


void SPA (stree * tree, POS * pos, u_int32_t phase, u_int32_t * extension, char *repeated_extension) {
	u_int32_t	rule_applied = 0;
	PATH		str;

	tree->e = phase + 1;	/* Leafs Trick: Apply implicit extensions 1 through prev_phase */

	/* Apply explicit extensions untill last extension of this phase is reached or extension rule 3 is applied once */
	for (; *extension <= phase+1; (*extension)++) {
		str.begin	= *extension;
		str.end		= phase + 1;

		SEA(tree, pos, str, &rule_applied, *repeated_extension);	/* Call Single-Extension-Algorithm */

		/* Check if rule 3 was applied for the current extension */
		if (rule_applied == 3) {
			/* Signaling that the next phase's first extension will not follow a suffix link because same extension is repeated */
			*repeated_extension = 1;
			break;
		}
		*repeated_extension = 0;
	}
	return;
}


stree *ST_AddStringList(stree *tree, lchar *str, u_int32_t length) {
	u_int32_t	phase, extension;
	char		repeated_extension = 0;
	POS		pos;

	if (str == 0) return(NULL);

	tree->length		= length;
	tree->tree_string	= str;
	ST_ERROR		= length + 10;

	tree->root->sons = create_node(tree, tree->root, 1, tree->length, 1);

	/* initialize algorithm parameters */
	extension	= 2;
	phase		= 2;
	suffixless	= NULL;
	pos.node	= tree->root;
	pos.edge_pos	= 0;

	/* Ukkonen's algorithm begins here */
	for (; phase < tree->length; phase++)
		SPA(tree, &pos, phase, &extension, &repeated_extension);

	return(tree);
}


stree *ST_CreateTree(void) {
	stree		*tree;

	/* Allocating the tree */
	if ((tree = calloc(1, sizeof(stree))) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	/* Allocating the tree root node */
	tree->root		= create_node(tree, 0, 0, 0, 0);
	tree->root->suffix_link	= NULL;

	return(tree);
}



void ST_DeleteSubTree (stnode * node) {
	if (node == 0) return;							/* Recoursion stoping condition */
	if (node->right_sibling != 0) ST_DeleteSubTree (node->right_sibling);	/* Recoursive call for right sibling */
	if (node->sons != 0) ST_DeleteSubTree (node->sons);			/* Recoursive call for first son */
	free (node);								/* Delete node itself, after its whole tree was deleted as well */
}


void ST_DeleteTree (stree * tree) {
	if (tree == 0) return;
	ST_DeleteSubTree(tree->root);
	free(tree->tree_string);
	free(tree->string_offset);
	free(tree);
}


int ST_PrintNode(stree *tree, stnode *node1, long depth, int cs_only) {
	stnode		*node2 = node1->sons;
	u_int32_t	d = depth, start = node1->edge_label_start, end, id;
	char		*byte;

	end = get_node_label_end (tree, node1);

	if ((!cs_only || node1->flag) && (depth > 0)) {
		/* Print the branches coming from higher nodes */
		while (--d) printf (" ");

		/* Print the node itself */
		if (node1->flag) printf("\033[1;32m");
		for (; start <= end; start++) {
			if (tree->tree_string[start] && !(tree->tree_string[start] & 0x000000ff)) {
				id = tree->tree_string[start] >> 8;
				printf("\033[1;31m[%u]\033[1;0m", id);
				break;
			} else {
				byte = (char *) &(tree->tree_string[start]);
				putchar(isprint(*byte) ? *byte : '.');
			}
		}
		if (node1->flag) printf("\033[1;0m");
		putchar('\n');
	}

	while (node2) {
		ST_PrintNode(tree, node2, depth+1, cs_only);
		node2 = node2->right_sibling;
	}

	return(0);
}


void ST_PrintTree(stree *tree, int cs_only) {
	ST_PrintNode(tree, tree->root, 0, cs_only);
}


int print_label(stree *t, stnode *n, avl_tree *string_ids, int depth) {
	char		*byte;
	int		d;
	u_int32_t	id, start, end;

	if (!t || !n) return(-1);

	d	= depth;
	id	= 0;
	start	= n->edge_label_start;
	end	= get_node_label_end(t, n);


	if (d) while (--d) printf(" ");
	// Print the node itself
	if (n->flag) printf("\033[1;32m");
	for (; start <= end; start++) {
		if (t->tree_string[start] && !(t->tree_string[start] & 0xffff00ff)) {
			id = t->tree_string[start] >> 8;
			printf("\033[1;31m[%u]\033[1;0m", id);
			break;
		} else {
			byte = (char *) &(t->tree_string[start]);
			putchar(isprint(*byte) ? *byte : '.');
		}
	}
	if (n->flag) printf("\033[1;0m");
	printf("\t");
	avl_subtree_print(string_ids->root);
	printf("\n");

	return(id);
}


void list_substrings(stree *t, stnode *cstr_node, ssize_t length) {
	u_int32_t	id, start, end, i;
	stnode		*n;

	id = 0;

	if (!cstr_node || !length) return;

	n = cstr_node;

	// DFS to find leaf nodes
	for (;;) {
		while (n->sons) n = n->sons;

		// now we're at a leaf so we found a substring
		start	= n->edge_label_start;
		end	= get_node_label_end(t, n);
		if (start) for (start = n->edge_label_start; start <= end; start++) {
			if (t->tree_string[start] && !(t->tree_string[start] & 0x000000ff)) {
				id = t->tree_string[start] >> 8;
				break;
			}
		}

		// print substring and meta infos
		start = n->path_position;
		printf("common substring starts in string \033[1;31m[%u]\033[1;0m\t at offset \033[1;33m%6u\033[1;0m  "
			"(absolute offset %7u) and has length \033[1;32m%3lu\033[1;0m (entropy: %f):\t'",
			id, start - t->string_offset[id-1]-1, start-1, (long unsigned int) length, entropy(t->tree_string+start, length));
		for (i=0; i<length; i++) putchar(isprint(t->tree_string[i+start]) ? t->tree_string[i+start] : '.');
		printf("'\n");

		if (!n->right_sibling) {
			n = n->father;
			while (n != cstr_node && !n->right_sibling) n = n->father;
			if (n == cstr_node) break;
		}
		n = n->right_sibling;
	}
	return;
}

void label(stree *t, stnode *n) {
	u_int32_t start, end, id;
	char *byte;

	start	= n->edge_label_start;
	end	= get_node_label_end(t, n);

	for (; start <= end; start++) {
		if (t->tree_string[start] && !(t->tree_string[start] & 0xffff00ff)) {
			id = t->tree_string[start] >> 8;
			printf("\033[1;31m[%u]\033[1;0m", id);
			break;
		} else {
			byte = (char *) &(t->tree_string[start]);
			putchar(isprint(*byte) ? *byte : '.');
		}
	}
}


ssize_t calc_common_subtree(stree *t, substr_list *cstr_list, ssize_t min_length, lcatbl *lca_table, stnode **leaves, ssize_t num_leaves, ssize_t num_of_files) {
	u_int32_t	id, num_cstr, left;
	ssize_t		i, nl;
	stnode		*lca, *tmp, **l;

	num_cstr	= 0;
	nl		= 0;
	left		= 0;

	if ((l = calloc(num_of_files, sizeof(stnode *))) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	t->root->flag = 1;	// always flag the root node

	// process leaves from left to right by sequentially moving left and right interval boundaries
	for (i=0; i<num_leaves; i++) {
		id = leaves[i]->string_id;	// get string ID of current leaf

		if (nl == 0) {
			// spool over equal string IDs to find left boundary
			while (i+1 < num_leaves && leaves[i+1]->string_id == leaves[i]->string_id) i++;

			// initialize left interval boundary
			left	= i;
		}

		if (l[id-1] == NULL) nl++;	// increase leaf counter if list entry for ID is empty
		l[id-1]	= leaves[i];		// update list entry so that it always contains the right-most leaf for this ID

		// check if left interval boundary must be moved
		if (left < i && id == leaves[left]->string_id) {
			left++;

			// spool over equal string IDs to find new left boundary
			while (left < i && leaves[left+1]->string_id == leaves[left]->string_id) {
				left++;
			}
			l[leaves[left]->string_id-1] = leaves[left];
		}

		if (nl == num_of_files) {
			// interval contains every ID at least one time, so it covers a common subtree

			lca = lca_lookup(lca_table, leaves[left], leaves[i]);
			if (!lca->flag) {
				num_cstr++;
				substr_add(cstr_list, lca, left, i, get_node_label_end(t, lca)-lca->path_position+1);
			}

			// flag all nodes on the (reverse) path from the root to the lca
			for (tmp = lca; !tmp->flag; tmp = tmp->father) tmp->flag = 1;

			if (i+1 < num_leaves && lca_lookup(lca_table, lca, leaves[i+1]) != lca) {
				// next leaf and current lca are in differen subtrees, so we can reset the ID interval
				nl = 0;
				memset(l, 0, num_of_files * sizeof(stnode *));
			} else {
				// remove ID corresponding to the left interval boundary from the interval
				l[leaves[left]->string_id-1]	= NULL;
				nl--;
				left++;

				// spool over equal string IDs to find new left boundary
				while (left < i && leaves[left+1]->string_id == leaves[left]->string_id) left++;
				l[leaves[left]->string_id-1] = leaves[left];
			}
		} else {
			if (lca_lookup(lca_table, leaves[left], leaves[i]) == t->root) {
				// interval does not cover a real subtree, so skip it
				memset(l, 0, num_of_files * sizeof(stnode *));

				// now restore list entry
				l[id-1]	= leaves[i];
				left = i;
				nl = 1;
			}
		}
	}
	free(l);

	return(num_cstr);
}
