/* trie.c
 *
 * Copyright (C) 2007 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "trie.h"


void print_trie(const trie t, int depth) {
	int i, j;
	for(i=0; i<t->childlist_len; i++) {
		if (t->childlist_len > 1) for (j=0; j<depth; j++) printf(" ");
		putchar(t->childlist[i].key);
		print_trie(&t->childlist[i], depth+1);
		if (t->childlist_len > 1) putchar('\n');
	}
	return;
}


/* find and return trie node for key */
trie trie_find_node(const trie t, const u_int16_t size, const u_char key) {
	u_int16_t i=0;
	u_int16_t high, low;
	
	if (t == NULL || size == 0) return(NULL);

	for (low=0, high=(size-1); high-low>0; ) {
		i = (high+low)/2;
		if (key <= t[i].key) high = i;
		else{
			i = (high+low+1)/2;
			low = i;
		}
	}
	return(t[i].key == key ? &t[i] : NULL);
}


/* do a lookup in a trie */
trie trie_find_memstr(trie t, const u_char* data, u_int16_t n) {
	u_int16_t k;
	trie node;

	node = t;
	for (k=0; k<n && node != NULL; k++) {
		node = trie_find_node(node->childlist, node->childlist_len, data[k]);
		if (node);// printf("%d: node is '%c'\n", k, node->key);
		else break;
	}
	
	return(node);
}


/* delete a path from a trie */
int trie_del_memstr(trie t, const u_char* data, u_int16_t n) {
	u_int16_t i = 0;
	u_int16_t high, low;
	trie node = t;

	if (!t || !data)	return(0);
	if (!t->childlist)	return(1);

	i = 0;
	for (low=0, high=(node->childlist_len-1); high-low>0; ) {
		i = (high+low)/2;
		if (*data <= node->childlist[i].key) high = i;
		else{
			i = (high+low+1)/2;
			low = i;
		}
	}

	if (node->childlist[i].key == data[0]) {
		/* check subtrie recursively */
		if (!trie_del_memstr(&node->childlist[i], &data[1], n-1)) return(0);

		/* if subpath was deleted, delete node from childlist */
		node->childlist_len--;
		if (node->childlist_len) {
			if (node->childlist_len > i) memmove(&(node->childlist[i]), &(node->childlist[i+1]), (node->childlist_len-i) * sizeof(trie_node));
			if ((node->childlist = realloc(node->childlist, node->childlist_len * sizeof(trie_node))) == NULL) {
				fprintf(stderr, "Error - Unable to reallocate memory: %s.\n", strerror(errno));
				exit(EXIT_FAILURE);
			}
			return(0);
		} else {
			free(node->childlist);
			node->childlist = NULL;
			return(1);
		}
	}

	return(0);
}


/* insert new node into trie with array nodes */
trie trie_insert_node(trie parent, const u_char key) {
	u_int16_t i = 0;
	int high, low;
	trie t = NULL;

	if ((t = realloc(parent->childlist, sizeof(trie_node)*(parent->childlist_len+1))) == NULL) {
		fprintf(stderr, "Error - Unable to allocate memory: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	parent->childlist = t;
	memset(&t[parent->childlist_len], 0, sizeof(trie_node));
	parent->childlist_len++;

	if (parent->childlist_len > 1) { // multiple nodes, have to memmove
		// binary search right position
		i = 0;
		for (low=0, high=(parent->childlist_len-1); high-low>0; ) {
			i = (high+low)/2;
			if (key <= t[i].key) high = i;
			else {
				i = (high+low+1)/2;
				low = i;
			}
		}
		if (key < t[i].key) memmove(&t[i+1], &t[i], (parent->childlist_len-i-1)*(sizeof(trie_node)));
		else  {
			if (i != parent->childlist_len-2) {
				memmove(&t[i+2], &t[i+1], (parent->childlist_len-i-1)*(sizeof(trie_node)));
			}
		}
		t = &t[i];
		
	}

	memset(t, 0, sizeof(trie_node));
	t->key = key;

	return(t);
}


/* inserts a binary string into a trie with array nodes */
trie trie_memins(trie t, const u_char *data, ssize_t n, void(*cbfn)(trie t)) {
	ssize_t k;
	trie target_node;

	if (t == NULL) {
		fprintf(stderr, "Error - Cannot insert into empty trie.\n");
		exit(EXIT_FAILURE);
	}

	for (k=0; k<n; k++) {
		/* search for key in child list */
		if ((target_node = trie_find_node(t->childlist, t->childlist_len, data[k])) == NULL) {
			if ((t = trie_insert_node(t, data[k])) == NULL) {
				fprintf(stderr, "Error - Unable to create new trie node: %s.\n", strerror(errno));
				exit(EXIT_FAILURE);
			} 
		} else t = target_node;
		if (cbfn) cbfn(t);
	}

	/* return node for last byte of data */
	return(t);
}


void trie_delete(trie t, u_int16_t node_size, void(*cbfn)(trie t)) {
	u_int16_t i = 0;

	if (t == NULL) return;

	for (i=0; i<node_size; i++) {
		if (cbfn) cbfn(&t[i]);
		trie_delete(t[i].childlist, t[i].childlist_len, cbfn);
	}
	free(t);

	return;
}
