/* util.c
 *
 * Copyright (C) 2007-2008 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as
 * published by the Free Software Foundation.  You may not use, modify or
 * distribute this program under any other version of the GNU General
 * Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "cluster.h"
#include "hash.h"
#include "nebula.h"
#include "queue.h"
#include "sig.h"
#include "signals.h"
#include "trie.h"
#include "util.h"


int regular_file(const struct dirent *e) {
	struct stat sta;
	memset(&sta, 0, sizeof(struct stat));

	if (stat(e->d_name, &sta) != 0) {
		fprintf(stderr, "Error - Unable to get file status for %s: %s.\n", e->d_name, strerror(errno));
		exit(EXIT_FAILURE);
	}

	return(S_ISREG(sta.st_mode));
}


int timesort(const void *a, const void *b) {
	struct dirent **A, **B;
	struct stat sta, stb;
	memset(&sta, 0, sizeof(struct stat));
	memset(&stb, 0, sizeof(struct stat));

	A = (struct dirent **) a;
	B = (struct dirent **) b;

	if ((stat((*A)->d_name, &sta) != 0) || (stat((*B)->d_name, &stb) != 0)) {
		fprintf(stderr, "Error - Unable to get file status for %s, %s: %s.\n", (*A)->d_name, (*B)->d_name, strerror(errno));
		exit(EXIT_FAILURE);
	}

	if (sta.st_ctime < stb.st_ctime) return(-1);
	if (sta.st_ctime > stb.st_ctime) return(1);
	return(0);
}


int idcmp(const void *a, ssize_t sizea, const void *b, ssize_t sizeb) {
	return(memcmp(a, b, sizea < sizeb ? sizea : sizeb));
}


inline double entropy(lchar *data, ssize_t len) {
	ssize_t		i;
	u_int32_t	bytefreqs[256];
	double		e;
	double		log2;

	e	= 0;
	log2	= log(2);	// factor for changing the log base

	// calculate byte frequencies
	memset(bytefreqs, 0, 256 * sizeof(u_int32_t));
	for (i=0; i<len; i++) bytefreqs[(u_char) data[i]]++;


	// calculate entropy
	for (i=0; i<=255; i++)
		if (bytefreqs[i])
			e -= (double)bytefreqs[i]/(double)len * log((double)bytefreqs[i]/(double)len) / log2;

	return(e);
}


void cleanup(void) {
	int i;

	if (!lock_mutex) return;
	lock_mutex = 0;

	// close listening socket
	shutdown(pfdset[0].fd, SHUT_RDWR);

	// close all open sessions
	for (i = 0; i <= POLLFD_MAX_SET_SIZE; i++) shutdown(pfdset[i].fd, SHUT_RDWR);

	// free job queues
	queue_free(siggenq, NULL);
	queue_free(classifyq, NULL);
	
	// free hash queues
	queue_free(clusterq, cluster_free);
	queue_free(outlierq, hash_free);

	// free tries
	trie_delete(md5sum_trie.childlist, md5sum_trie.childlist_len, NULL);
	trie_delete(spamsum_trie.childlist, spamsum_trie.childlist_len, NULL);
	trie_delete(sighash_trie.childlist, sighash_trie.childlist_len, NULL);

	// destroy semaphores
	if (repsem_getvalue(&sessions_semaphore, &i) == -1) {
		fprintf(stderr, "Error - Unable to get semaphore value: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	for (; i; i--) repsem_wait(&sessions_semaphore);
	if (repsem_destroy(&sessions_semaphore) == -1) {
		fprintf(stderr, "Error - Unable to destroy semaphore: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	if (repsem_getvalue(&siggen_semaphore, &i) == -1) {
		fprintf(stderr, "Error - Unable to get semaphore value: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	for (; i; i--) repsem_wait(&siggen_semaphore);
	if (repsem_destroy(&siggen_semaphore) == -1) {
		fprintf(stderr, "Error - Unable to destroy semaphore: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	// destroy mutexes and rwlocks
	pthread_mutex_destroy(&siggen_mutex);
	pthread_mutex_destroy(&sigwrite_mutex);

	pthread_rwlock_destroy(&md5sum_trie_lock);
	pthread_rwlock_destroy(&spamsum_trie_lock);
	pthread_rwlock_destroy(&sighash_trie_lock);
	pthread_rwlock_destroy(&sidlock);

	return;
}
